/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description: 
 *		Bulk pipe routines for the tcan_usb driver
 *
 * TODO:
 *
 * History of changes:
 *	 2005/06/05 - 0.2 - minor fixes for the send-thread
 *	 2005/04/30 - 0.1.1 - fixed incorrect messagecount given by usbd (bug check)
 *	 2005/03/15 - 0.1 - first version
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


#include "tcan_usb.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_DeleteBulkResources)
#endif 


/******************************************************************************
 
   Tc_StopPolling

 Description:
	Stop the polling irp by cancel it.
	isInterrupt determines either an interrupt or the read irp.
	Do this under protection of a spin lock.
	

  @devExt - the device Extension
  @isInterrupt

 no return value

******************************************************************************/
VOID Tc_StopPolling
(
 IN PDEVICE_EXTENSION	devExt,
 IN BOOLEAN				isInterrupt
)
{
	KdPrint (( DRVNAME "Tc_StopPolling() - %s\n", ( isInterrupt ? "Interrupt" : "Read" ) ));
	if ( isInterrupt)
	{
		if ( devExt->IntPollPending )
		{
			IoCancelIrp ( devExt->IntIrp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, devExt->IntIrp );
		}
	}
	else
	{
		if ( devExt->ReadPollPending )
		{
			IoCancelIrp ( devExt->ReadPollIrp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, devExt->ReadPollIrp );
		}
	}
}


/******************************************************************************
 
   Tc_DeleteBulkResources

 Description:
	free memory ressources of urb and irp allocated by StartDevice	

  @devExt - the device Extension

 no return value

******************************************************************************/
VOID Tc_DeleteBulkResources
(
 IN PDEVICE_EXTENSION devExt
)
{
	PAGED_CODE ();
	if ( devExt->ReadPollUrb )
	{
		ExFreePool ( devExt->ReadPollUrb );
		devExt->ReadPollUrb = NULL;
	}

	if ( devExt->ReadPollIrp )
	{
		IoFreeIrp ( devExt->ReadPollIrp );
		devExt->ReadPollIrp = NULL;
	}
}


/******************************************************************************
 
   Tc_StartPolling

 Description:
	As one of the steps of CAN_START
	This routine starts a pending read urb to read the bulk in endpoint

  @devExt - the device Extension
  @isInterrupt

 returns a ntStatus

******************************************************************************/
NTSTATUS Tc_StartPolling
(
 IN PDEVICE_EXTENSION	devExt,
 IN BOOLEAN				isInterrupt
)
{	
	PIRP				Irp;
	PURB				Urb;
	KIRQL				iLevel;
	PVOID				urbBuffer;
	NTSTATUS			ntStatus;
	KLOCK_QUEUE_HANDLE	hLock;
	PIO_STACK_LOCATION	ioStack;

	KdPrint((DRVNAME "POLL: Tc_Start%sUrb... at %X\n", ( isInterrupt ? "Interrupt" : "Read" ), KeGetCurrentIrql() ));

	if ( isInterrupt )
	{
		KeAcquireInStackQueuedSpinLock ( &devExt->InterruptLock, &hLock );
		/*----------------------------------------------------------------*/
			if ( devExt->IntPollPending )
			{
				KeReleaseInStackQueuedSpinLock ( &hLock );
				KdPrint (( DRVNAME "POLL: Failed; Interrupt poll is pending\n" ));
				return STATUS_DEVICE_BUSY;
			}
			else
				devExt->IntPollPending = TRUE;
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );
	}
	else
	{
		KeAcquireInStackQueuedSpinLock ( &devExt->ReadPollLock, &hLock );
		/*----------------------------------------------------------------*/
			if (devExt->ReadPollPending)
			{
				KeReleaseInStackQueuedSpinLock ( &hLock );
				KdPrint (( DRVNAME "POLL: Failed; Read poll is pending\n" ));
				return STATUS_DEVICE_BUSY;
			}
			else
				devExt->ReadPollPending = TRUE;
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );
	}
	/*KdPrint((DRVNAME "POLL: 1\n" ));*/

	Irp = ( isInterrupt ? devExt->IntIrp : devExt->ReadPollIrp );
	Urb = ( isInterrupt ? devExt->IntUrb : devExt->ReadPollUrb );
	ASSERT ( Irp && Urb );

	ntStatus 
		= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );
	
	/*KdPrint((DRVNAME "POLL: 2\n" ));*/
	if ( !NT_SUCCESS(ntStatus) )
	{
		if ( isInterrupt )
			devExt->IntPollPending = FALSE;
		else
			devExt->ReadPollPending = FALSE;
		KdPrint (( DRVNAME "POLL: Failed to acquire remove lock\n" ));
		return ntStatus;
	}

	Irp->Cancel = FALSE;

	if ( !isInterrupt )
		RtlZeroMemory ( &devExt->ReadPollBuffer, sizeof(BULK_BUFFER) );

	/* Initialize the URB we use for reading the interrupt pipe */
	if ( isInterrupt )
		urbBuffer = &devExt->IntBuffer;
	else
		urbBuffer = &devExt->ReadPollBuffer;

	UsbBuildInterruptOrBulkTransferRequest
	(
		Urb, 
		sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
		( isInterrupt ? devExt->InterruptINPipe : devExt->BulkINPipe ), 
		urbBuffer, 
		NULL, 
		( isInterrupt ? sizeof(UCI_INT_DATA) : sizeof(BULK_BUFFER) ), 
		USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK, 
		NULL
	);

	/*KdPrint((DRVNAME "POLL: 3\n" ));*/
	/* 
	 *  set the processing handler as completion routine
	 */
	IoSetCompletionRoutine ( 
		Irp, 
		(PIO_COMPLETION_ROUTINE) ( isInterrupt ? Tc_OnInterrupt : Tc_OnBulkRead ), 
		devExt, 
		TRUE, 
		TRUE, 
		TRUE );

	ioStack
		= IoGetNextIrpStackLocation ( Irp );

	ioStack->MajorFunction
		= IRP_MJ_INTERNAL_DEVICE_CONTROL;

	ioStack->Parameters.DeviceIoControl.IoControlCode
		= IOCTL_INTERNAL_USB_SUBMIT_URB;

	ioStack->Parameters.Others.Argument1
		= Urb;

	KdPrint (( DRVNAME "POLL: Tc_StartPolling...ntStatus(%X)\n", ntStatus ));

	/*
	 *  Issue the Interrupt and Read Polling at DISPATCH_LEVEL
	 *	This is only for a consequent line of level order. This routine is expected to be executed at DISPATCH_LEVEL
	 */
	KeRaiseIrql ( DISPATCH_LEVEL,  &iLevel );
	ntStatus = IoCallDriver ( devExt->NextDeviceObject, Irp );
	KeLowerIrql ( iLevel );
	KdPrint (( DRVNAME "POLL: Tc_StartPolling...end\n" ));
	return ntStatus;
}


/******************************************************************************
 
   Tc_OnBulkRead

 Description:
	This is the completion routine for a read operation on the interrupt pipe.

 Execution Level: 
	DISPATCH_LEVEL

  @deviceObject - the device object
  @Irp
  @devExt

 returns a ntStatus

******************************************************************************/
NTSTATUS Tc_OnBulkRead
(
 IN PDEVICE_OBJECT		deviceObject, 
 IN PIRP				Irp, 
 IN PDEVICE_EXTENSION	devExt
)
{
	BULK_BUFFER Buffer;
	KLOCK_QUEUE_HANDLE hLock;

	KdPrint (( DRVNAME "READPOLL: Tc_OnBulkRead...level(%i)\n", KeGetCurrentIrql() ));

	ASSERT ( KeGetCurrentIrql() <= DISPATCH_LEVEL );

	Buffer = devExt->ReadPollBuffer;

	KeAcquireInStackQueuedSpinLock ( &devExt->ReadPollLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->ReadPollPending = FALSE;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	KdPrint (( DRVNAME "READPOLL: --- Bulk IN Message arrived with Cancel Flag %i ! -----\n", Irp->Cancel ));
	if ( NT_SUCCESS(Irp->IoStatus.Status) && !Irp->Cancel && devExt->CanStarted )
	{
		PUCI_CANMSG pMsg;
		PBULK_FIFO	RXFifo;

		KdPrint (( DRVNAME "READPOLL: --- Bulk IN Message arrived ! -----\n" ));
		InterlockedExchange ( &devExt->cacheFWOvr, 0 );
		/* Buffer.message1.msgdata[7] = 0x0;
		Buffer.message2.msgdata[7] = 0x0;
		Buffer.message3.msgdata[7] = 0x0;
		Buffer.message4.msgdata[7] = 0x0;
		KdPrint((DRVNAME "number of messages: %X\n", Buffer.mcount));
		KdPrint((DRVNAME "message1: %s\n", Buffer.message1.msgdata));
		KdPrint((DRVNAME "message2: %s\n", Buffer.message2.msgdata));
		KdPrint((DRVNAME "message3: %s\n", Buffer.message3.msgdata));
		KdPrint((DRVNAME "message4: %s\n", Buffer.message4.msgdata));*/
		
		RXFifo = &devExt->RXFifo;
		pMsg = &Buffer.message1;

		if ( Buffer.mcount <= BULK_MSG_COUNT )
		{
			/*  lock the receive fifo
			 *
			 *  a mutex is not useable because this completion routine is mostly called at dispatch level
			 */
			USHORT i;
			KeAcquireInStackQueuedSpinLock ( &devExt->RXFifoLock, &hLock );
			/*----------------------------------------------------------------*/
				for ( i = 0; i < Buffer.mcount; i++ )
				{
					ULONG			tail;
					LARGE_INTEGER	t;
					char			tmp[CAN_MSG_LENGTH];

					KdPrint (( DRVNAME "READPOLL: --- Msg store loop %i -----\n", i ));
					if ( RXFifo->status & CANMSG_BUF_FULL )
					{
						RXFifo->status |= CANMSG_BUF_OVR;
						/*  we need to notify the application about this fifo overrun to  */

						InterlockedIncrement ( &devExt->RXFifoOverrun );
						
						if ( devExt->IntEvent && !devExt->IntSet )
						{
							KLOCK_QUEUE_HANDLE hLockLock;
							KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLockLock );
							/*----------------------------------------------------------------*/							
								RtlZeroMemory ( &devExt->IntEventBuffer, sizeof(UCI_INT_DATA) );
								devExt->IntEventBuffer.st_code = 0x90;
								devExt->IntEventBuffer.overruns = (unsigned char) devExt->RXFifoOverrun;
								devExt->IntSet = TRUE;

								if ( devExt->IoMode )
								{
									KeSetEvent ( devExt->IntEvent, EVENT_INCREMENT, FALSE );
								}
								else
								{
									/* Start a dpc for processing an interrupt Irp */
									KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) FALSE );
								}
							/*----------------------------------------------------------------*/
							KeReleaseInStackQueuedSpinLock ( &hLockLock );
						}
						break;
					}
					tail = RXFifo->tail;
					RXFifo->msg[tail].flags = pMsg->flags;
					if ( pMsg->flags & CAN_EFF )
						RXFifo->msg[tail].id = (pMsg->id1 << 21)|(pMsg->id2 << 13)
							|(pMsg->id3 << 5)|(pMsg->id4 >> 3);
					else
						RXFifo->msg[tail].id = (pMsg->id1 << 3)|(pMsg->id2 >> 5);
					RtlZeroMemory ( RXFifo->msg[tail].msgdata, CAN_MSG_LENGTH );
					RtlCopyMemory ( RXFifo->msg[tail].msgdata, pMsg->msgdata, pMsg->flags & CAN_DLC );
					
					KeQuerySystemTime ( &t );
					ExSystemTimeToLocalTime ( &t, &RXFifo->msg[tail].rtime );

					KdPrint (( DRVNAME "READPOLL: --- Message stored in fifo ! -----\n" ));
					
					RtlZeroMemory ( tmp, CAN_MSG_LENGTH );
					RtlCopyMemory ( tmp, RXFifo->msg[tail].msgdata, CAN_MSG_LENGTH-1 );
					KdPrint (( DRVNAME "READPOLL: id: %i, %s, tail: %i\n", pMsg->id1, tmp, tail ));

					RXFifo->tail++;
					pMsg++;
					if ( RXFifo->tail == devExt->RXFifoSize ) RXFifo->tail = 0;
					if ( RXFifo->tail == RXFifo->head ) RXFifo->status |= CANMSG_BUF_FULL;
				}
				RXFifo->status |= CANMSG_BUF_PEND;
			/*----------------------------------------------------------------*/
			KdPrint (( DRVNAME "READPOLL: --- lock release ! -----\n" ));
			KeReleaseInStackQueuedSpinLock ( &hLock );
		}
		else
		{
			/*
			 *  if needed we could send an interrupt to the application
			 *  currently there's no need for
			 */
			TRUE;
			KdPrint (( DRVNAME "READPOLL: --- INVALID MESSAGE ! -----\n" ));
		}
		KdPrint (( DRVNAME "READPOLL: --- before start polling ! -----\n" ));
		Tc_StartPolling ( devExt, FALSE );

		if ( !devExt->ReadPending )
		{
			if ( devExt->IoMode )
			{
				/* we need this spin lock, to prevent something from dereference the event, while we are signaling it */
				KeAcquireInStackQueuedSpinLock ( &devExt->ReadEventLock, &hLock );
				/*----------------------------------------------------------------*/
					if ( devExt->ReadEvent && !devExt->ReadCount ) 
					{
						KdPrint (( DRVNAME "OnBulkRead setevent, readcount(%i)\n", devExt->ReadCount ));
						KeSetEvent ( devExt->ReadEvent, EVENT_INCREMENT, FALSE );
					}
					else
						KdPrint (( DRVNAME "OnBulkRead event not set, readcount(%i)\n", devExt->ReadCount ));
				/*----------------------------------------------------------------*/
				KeReleaseInStackQueuedSpinLock ( &hLock );
			}
			else
			{
				/* Start a dpc for processing a read Irp */
				KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) TRUE );
				/* KdPrint (( DRVNAME "about to start a dpc at %X\n", KeGetCurrentIrql() )); */
			}
		}
	}
	else
	{
		KdPrint (( DRVNAME "READPOLL: IRP canceled; ntStatus(%X); UrbStatus(%X)\n",
			Irp->IoStatus.Status, URB_STATUS(devExt->IntUrb) ));
	}

	IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );

	return STATUS_MORE_PROCESSING_REQUIRED;
}


/******************************************************************************
 
   Tc_DispatchReadWrite

 Description:
	This routine handles the user requests for read / write.
	Several validations such as the buffer size or the queue state must be made.
	If a request for the desired type is being processed, it queues the request.

 Execution Level:
	PASSIVE_LEVEL by the Dispatcher

  @deviceObject - the device object
  @Irp

 returns a ntStatus

******************************************************************************/
NTSTATUS Tc_DispatchReadWrite
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
)
{
	ULONG				size;
	KIRQL				iLevel;
	BOOLEAN				slRelease;
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	KLOCK_QUEUE_HANDLE	hLock;
	PIO_STACK_LOCATION	ioStack;

	KdPrint((DRVNAME "Tc_DispatchReadWrite with (%X)!\n",KeGetCurrentIrql()));

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	if ( !devExt->CanStarted )
	{
		/*KdPrint (( DRVNAME "FAILED; Can device is stopped\n" ));*/
		return Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
	}
	if ( devExt->DevState == Removed )
	{
		/*KdPrint (( DRVNAME "FAILED; queue inactive, device removed\n" ));*/
		return Tc_FinishIrp2 ( Irp, STATUS_DEVICE_REMOVED );
	}

	/*
	 *	Direct IO
	 */
	size 
		= Irp->MdlAddress ?  MmGetMdlByteCount ( Irp->MdlAddress ) : 0;
	
	ioStack = IoGetCurrentIrpStackLocation ( Irp );	

	/* ----------- obsolete ----------------
	 *	Buffered IO
	 *
	 * ULONG size 
	 *	= ioStack->Parameters.Read.Length;
	   ----------- obsolete ---------------- */

	if ( size < sizeof(CANMSG) )
	{
		/* maybe STATUS_INVALID_PARAMETER is the better choice */
		KdPrint (( DRVNAME "FAILED; invalid buffer size for can message; to small\n" ));
		return Tc_FinishIrp2 ( Irp, STATUS_BUFFER_TOO_SMALL );
	}

	ntStatus 
		= STATUS_PENDING;
	slRelease = TRUE;

	/*
	 * The dispatch routine runs under protection of a spin lock for IoQueueState,
	 * because of the case the queue gets InActive and we start processing the irp.
	 * It happens that the irp before submits the stop event for our waiting pnp
	 * request, thus the pnp-management is in an invalid state.
	 */
	KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
	iLevel = hLock.OldIrql;
	/*----------------------------------------------------------------*/	
		/*
		 *  if the io-queue is stalled, the best thing to do is to
		 *  insert the irp into the queue.
		 */
		if ( devExt->IoQueueState == Stalled )
			IoCsqInsertIrp ( &devExt->IoQueue, Irp, NULL );
		else
		if ( devExt->IoQueueState == InActive || devExt->DevState == Stopped )
		{
			/*KdPrint((DRVNAME "FAILED; queue inactive, device removed\n"));*/
			ntStatus = Tc_FinishIrp2 ( Irp, STATUS_DEVICE_NOT_READY );
		}
		else	/* Io-Queue must be Active */
		{
			/*
			 * Check whether a read or write is pending
			 *		do this under protection of a spin lock, 
			 *		because of a potential race condition with the completion routine
			 */
			KLOCK_QUEUE_HANDLE hLockLock;
			if ( ioStack->MajorFunction == IRP_MJ_WRITE )
			{
				IoCsqInsertIrp ( &devExt->IoQueue, Irp, NULL );
				KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, FALSE );
			}
			else
			{
				KeAcquireInStackQueuedSpinLock ( &devExt->ReadLock, &hLockLock );
				/*KdPrint (( DRVNAME "Status %X\n", devExt->RXFifo.status ));*/
				/*----------------------------------------------------------------*/
					if ( devExt->ReadPending || !devExt->RXFifo.status )
					{
						IoCsqInsertIrp ( &devExt->IoQueue, Irp, NULL );
						/* KdPrint (( DRVNAME "Inserted in queue\n" )); */
						/*----------------------------------------------------------------*/
						KeReleaseInStackQueuedSpinLock ( &hLockLock );
					}
					else
					{
						devExt->ReadPending = TRUE;
						/*----------------------------------------------------------------*/
						KeReleaseInStackQueuedSpinLock ( &hLockLock );
						hLock.OldIrql = DISPATCH_LEVEL;
						KeReleaseInStackQueuedSpinLock ( &hLock );
						slRelease = FALSE;
						IoMarkIrpPending ( Irp );
						Tc_StartReadIo ( deviceObject, Irp );
						KeLowerIrql ( iLevel );
					}
			}			
		}

	/*----------------------------------------------------------------*/
    if ( slRelease )
		KeReleaseInStackQueuedSpinLock ( &hLock );

	return ntStatus;
}


/******************************************************************************
 
   Tc_StartIoDpc

 Description:

  This routine processes the io-queue for eiter read or write irps. 
  If there're no irps in the queue, it ends.
  it checks the queue and device state and calls StartXxxIo routines
  if it needs to.	

 Execution IRQL:
	DISPATCH_LEVEL

  @Dpc
  @deviceObject - the device object
  @Irp
  @isRead - signals either a read or a interrupt irp to process on

 no return value

******************************************************************************/
VOID Tc_StartIoDpc
(
 IN PKDPC				Dpc,
 IN	PDEVICE_OBJECT		deviceObject,
 IN	PIRP				pIrp,
 IN BOOLEAN				isRead
)
{
	PDEVICE_EXTENSION devExt;

	KdPrint (( DRVNAME "DPC: ---+++ Tc_StartIoDpc start at level (%X) +++-----!\n", KeGetCurrentIrql() ));

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	if ( !devExt->CanStarted || devExt->DevState == Removed || devExt->DevState == Stopped || devExt->DevState == SurpriseRemoved )
	{
		KdPrint (( DRVNAME "DPC: FAILED; Can device is stopped or removed\n" ));
	}
	else
	{
		NTSTATUS			ntStatus;
		KLOCK_QUEUE_HANDLE	hLock;
		ntStatus 
			= STATUS_PENDING;
		/*
		 * The dispatch routine runs under protection of a spin lock for IoQueueState,
		 * because of the case the queue gets InActive and we start processing the irp.
		 * It happens that the irp before submits the stop event for our waiting pnp
		 * request, thus the pnp-management is in an invalid state.
		 */
		KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
		/*----------------------------------------------------------------*/
			/*
			 * devExt->CanStarted could be changed while we're spinning for the queue state lock
			 */
			if ( devExt->IoQueueState == Active && devExt->CanStarted )
			{
				/*
				 * Check whether a read or interrupt is pending
				 *		do this under protection of a spin lock, 
				 *		because of a potential race condition with the completion routine
				 */
				PIRP Irp;
				QUEUECONTEXT PeekContext;
				PeekContext.Context = ( isRead ? &devExt->Read : &devExt->Interrupt );
				
				Irp
					= IoCsqRemoveNextIrp ( &devExt->IoQueue, &PeekContext );

				if ( Irp )
				{
					/*----------------------------------------------------------------*/
					KeReleaseInStackQueuedSpinLock ( &hLock );
					if ( isRead )
					{
						Tc_StartReadIo ( deviceObject, Irp );
						return;
					}
					else
					{
						Tc_StartInterruptIo ( deviceObject, Irp );
						return;
					}
				}
			}
			/* Io-Queue state must be Stalled or InActive */
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );
	}

	if ( isRead )
	{
		KLOCK_QUEUE_HANDLE hLockLock;
		KeAcquireInStackQueuedSpinLock ( &devExt->ReadLock, &hLockLock );
		/*----------------------------------------------------------------*/
			devExt->ReadPending = FALSE;
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLockLock );

		KeSetEvent ( &devExt->ReadStopEvent, 0, FALSE );
	}
}


/******************************************************************************
 
   Tc_StartReadIo

 Description:

  This routine processes the io-queue for all read irps. If there's no read
  irp in the queue, it ends.
    
  Irp comes either from DispatchReadWrite or StartIoDpc.

 Note: The try expressions should made as performant as possible, because it
	   locks the RX-Fifo for processing which is also used by the read
	   polling irp. Read the firmware's buffer has the higher prio.
	
 Execution IRQL:
	DISPATCH_LEVEL

  @deviceObject - the device object
  @Irp

 no return value

******************************************************************************/
VOID Tc_StartReadIo
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	PDEVICE_EXTENSION devExt;

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	KdPrint (( DRVNAME "READ: ---+++ Tc_StartReadIo start at level (%X) +++-----!\n", KeGetCurrentIrql() ));

	/*
	 *  if the can device is deactivated meanwhile we fail the irp and sends an stop event
	 */
	if ( !devExt->CanStarted )
		Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
	else
	{
		NTSTATUS ntStatus;
		ntStatus = STATUS_UNSUCCESSFUL;

		Irp->IoStatus.Information = 0;
		
		ntStatus
			= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );
		
		if ( !NT_SUCCESS(ntStatus) )
			Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
		else
		{
			ULONG				size;
			ULONG				mcount;
			ULONG				headToEnd;
			BOOLEAN				forward;
			ULONG_PTR			Buffer;
			KLOCK_QUEUE_HANDLE	hLock;

			__try
			{
				PBULK_FIFO RXFifo = &devExt->RXFifo;

				ASSERT ( KeGetCurrentIrql() <= DISPATCH_LEVEL );

				/*
				 *	Direct IO
				 */
				size 
					= Irp->MdlAddress ?  MmGetMdlByteCount ( Irp->MdlAddress ) : 0;				
				if (!size)
				{
					KdPrint (( DRVNAME "READ: this should not be possible; the dispatch routine caught this !\n" ));
					ntStatus = STATUS_INVALID_PARAMETER;
					__leave;
				}

				/*
				 *	Direct IO
				 *	WIN9X: does only support MmGetSystemAddressForMdlSafe
				 */
				Buffer 
					= (ULONG_PTR) MmGetSystemAddressForMdlSafe ( Irp->MdlAddress, NormalPagePriority );				
				if ( !Buffer )
				{
					KdPrint (( DRVNAME "READ: FAILED because of low ressources\n" ));
					ntStatus = STATUS_INSUFFICIENT_RESOURCES;
					__leave;
				}

				KeAcquireInStackQueuedSpinLock ( &devExt->RXFifoLock, &hLock );
				/*----------------------------------------------------------------*/

					/* we need to ensure that there are messages in the rxbuffer for this read operation */
					if ( !(RXFifo->status & CANMSG_BUF_FULL) && RXFifo->head == RXFifo->tail )
					{
						KeReleaseInStackQueuedSpinLock ( &hLock );
						KdPrint (( DRVNAME "READ: StartReadIo failed because of no data in the fifo\n" ));
						ntStatus = STATUS_NO_DATA_DETECTED;
						__leave;
					}

					forward 
						= RXFifo->head < RXFifo->tail;

					mcount 
						= ( !forward ? devExt->RXFifoSize-RXFifo->head+RXFifo->tail : RXFifo->tail - RXFifo->head );

					KdPrint (( DRVNAME "READ: mcount0: %i; size: %i; head: %i; tail: %i\n", mcount, size, RXFifo->head, RXFifo->tail ));
					
					mcount 
						= ( ( size/sizeof(CANMSG) ) < mcount ? size/sizeof(CANMSG) : mcount );

					headToEnd 
						= devExt->RXFifoSize-RXFifo->head;

					KdPrint((DRVNAME "READ: forward: %i; mcount: %i; headToEnd: %i\n", forward, mcount, headToEnd));
					KdPrint((DRVNAME "READ: head id: %i\n", RXFifo->msg[RXFifo->head].id));

					if ( forward || mcount <= headToEnd )
					{
						RtlCopyMemory ( (void *) Buffer, &(RXFifo->msg[RXFifo->head]), mcount*sizeof(CANMSG) );
						RXFifo->head += mcount;
					}
					else
					{
						RtlCopyMemory ( (void *) Buffer, &(RXFifo->msg[RXFifo->head]), headToEnd*sizeof(CANMSG) );
						RtlCopyMemory ( (void *)(((PCANMSG) Buffer)+headToEnd), RXFifo->msg, (mcount-headToEnd)*sizeof(CANMSG) );
						RXFifo->head = mcount-headToEnd;
					}

					RXFifo->status &= ~(CANMSG_BUF_FULL | CANMSG_BUF_OVR);
					
					/*
					 * if the fifo is not empty and there are no more read irps pending
					 * then set an read event
					 */			
					if ( RXFifo->tail == RXFifo->head ) 
						RXFifo->status = CANMSG_BUF_EMPTY;
					else
					{
						KLOCK_QUEUE_HANDLE hLockLock;
						KdPrint (( DRVNAME "READ: readcount (%X)\n", devExt->ReadCount ));
						
						KeAcquireInStackQueuedSpinLock ( &devExt->ReadEventLock, &hLockLock );
						/*----------------------------------------------------------------*/
							if ( devExt->ReadEvent && !devExt->ReadCount )
							{
								KdPrint (( DRVNAME "READ: setevent!\n" ));
								KeSetEvent ( devExt->ReadEvent, EVENT_INCREMENT, FALSE );
							}
							else
								KdPrint (( DRVNAME "READ: event not set\n" ));
						/*----------------------------------------------------------------*/
						KeReleaseInStackQueuedSpinLock ( &hLockLock );
					}
				/*----------------------------------------------------------------*/
				KeReleaseInStackQueuedSpinLock ( &hLock );
				ntStatus = STATUS_SUCCESS;
				Irp->IoStatus.Information = mcount * sizeof(CANMSG);
			}
			__finally
			{
				Tc_FinishIrp0 ( Irp, ntStatus );
				IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			}
		}
	}
	KdPrint (( DRVNAME "READ: ---+++ Tc_StartReadIo end +++-----!\n" ));

	KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) TRUE );
	return;
}


/******************************************************************************
 
   Tc_StartWriteIoThread

 Description:
	This Thread is responsible for the write operation. Because we can not
	block a thread at a level higher than passive level, we need to ensure
	that anytime we are processing a write operation is done at passive level.
	Beside this, queue a work item for doing this is not the intention of
	work items, though they were used by the system.


 Execution IRQL:
	PASSIVE_LEVEL

  @Context - context

 no return value

******************************************************************************/
VOID Tc_StartWriteIoThread
(
 IN PVOID Context
)
{
	NTSTATUS ntStatus;
	PDEVICE_EXTENSION devExt;
    QUEUECONTEXT PeekContext;
	PDEVICE_OBJECT deviceObject;
	LARGE_INTEGER StalledWaitInterval;

	KdPrint (( DRVNAME "WRITE: Write Thread created\n" ));

    deviceObject 
		= (PDEVICE_OBJECT) Context; 

    devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	PeekContext.Context = &devExt->Write;

    StalledWaitInterval.QuadPart = Int32x32To64 ( 250*1000 , -10 ); /* 250 ms */
    KeSetPriorityThread ( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

    while ( TRUE )
    {
		PIRP Irp = NULL;
		KLOCK_QUEUE_HANDLE hLock;
		KLOCK_QUEUE_HANDLE hLockLock;

        KeWaitForSingleObject (
			&devExt->WriteThreadSemaphore,
            Executive,
            KernelMode,
            FALSE,
            NULL );

		/*
		 * if the queue is stalled, we need to wait for a while and test it again,
		 * means, after waiting, increase the Semaphore and continue the loop.
		 * if the queue is inactive we need to terminate this thread
		 */
		if ( devExt->IoQueueState == Stalled )
		{
			KeSetEvent ( &devExt->WriteStopEvent, 0, FALSE );
			KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, FALSE );
			KeDelayExecutionThread ( KernelMode, FALSE, &StalledWaitInterval );
			continue;
		}
        if ( devExt->WriteThreadStop || devExt->IoQueueState == InActive ) 
		{
			KeSetEvent ( &devExt->WriteStopEvent, 0, FALSE );
			PsTerminateSystemThread ( STATUS_SUCCESS );
		}
		
		Irp
			= IoCsqRemoveNextIrp ( &devExt->IoQueue, &PeekContext );

        if ( !Irp )
		{
            KdPrint(("WRITE: seems something has somehow canceled an irp...\n"));
            continue;
        }

		KeAcquireInStackQueuedSpinLock ( &devExt->WriteLock, &hLock );
		/*----------------------------------------------------------------*/
			devExt->WritePending = TRUE;
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );

        KdPrint (( DRVNAME "WRITE: ---+++ StartWriteIoThread start at level (%X), QueueCount (%i) +++-----!\n", KeGetCurrentIrql(), devExt->ReadCount ));

		/*
		 * if the can device is deactivated meanwhile we fail the irp and sends an stop event
		 */
		if ( !devExt->CanStarted )
			Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
		else
		{
			ntStatus = STATUS_UNSUCCESSFUL;

			Irp->IoStatus.Information = 0;
			
			ntStatus
				= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );
			
			if ( !NT_SUCCESS(ntStatus) )
				Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
			else
			{
				ULONG size;
				ULONG msgcnt;
				PCANMSG pMsg;
				ULONG mcount;
				__try
				{
					size
						= Irp->MdlAddress ?  MmGetMdlByteCount ( Irp->MdlAddress ) : 0;

					mcount = size / sizeof(CANMSG);
					
					KdPrint (( DRVNAME "WRITE: mcount: %i; size: %i\n", mcount, size ));
					/*
					 *	Direct IO
					 *	WIN9X: does only support MmGetSystemAddressForMdl
					 */
					pMsg
						= (PCANMSG) MmGetSystemAddressForMdlSafe ( Irp->MdlAddress, NormalPagePriority );
					if ( !pMsg )
					{
						KdPrint (( DRVNAME "WRITE: FAILED because of low ressources\n" ));
						ntStatus = STATUS_INSUFFICIENT_RESOURCES;
						__leave;
					}

					/*KdPrint (( DRVNAME "WRITE: 1\n" ));*/
					/*------------------- obsolete -----------------------------
					 	Buffered IO
					
					    PCANMSG pMsg = (PCANMSG) Irp->AssociatedIrp.SystemBuffer;
					  ------------------- obsolete ----------------------------- */

					for ( msgcnt = 0; msgcnt < mcount; )
					{
						URB urb;
						ULONG i;
						CHAR dbgbuf[10];
						UCHAR buffer[64];
						ULONG transfered;
						PUCI_CANMSG pUciMsg;
						RtlZeroMemory ( buffer, 64 );
						
						pUciMsg 
							= (PUCI_CANMSG) (buffer+1);

						/*KdPrint (( DRVNAME "WRITE: 2, %i\n", msgcnt ));*/

						for ( i = 0; i < BULK_MSG_COUNT; i++ )
						{
							if ( msgcnt == mcount )
								break;

							/*KdPrint (( DRVNAME "WRITE: 2.5, %i\n", i ));*/

							pUciMsg->flags
								= (UCHAR) pMsg->flags & 0xff;

							/*KdPrint (( DRVNAME "WRITE: 3, %i\n", i ));*/

							if ( pMsg->flags & CAN_EFF ) 
							{ 
								/* ext msg frame */
								pUciMsg->id1 =  (UCHAR)( pMsg->id >> 21 );
								pUciMsg->id2 =  (UCHAR)( pMsg->id >> 13 );
								pUciMsg->id3 =  (UCHAR)( pMsg->id >> 5 );
								pUciMsg->id4 =  ( (UCHAR)( pMsg->id << 3 ) & 0xf8 );
							}
							else 
							{ 
								/* std msg frame */
								pUciMsg->id1 =  (UCHAR)( pMsg->id >> 3 );
								pUciMsg->id2 =  ( (UCHAR)( pMsg->id << 5 ) & 0xe0 );
							}
							/*KdPrint (( DRVNAME "WRITE: 4\n" ));*/
							RtlCopyMemory ( pUciMsg->msgdata, pMsg->msgdata, pMsg->flags & CAN_DLC );
							
							RtlZeroMemory ( dbgbuf, 10 );
							RtlCopyMemory ( dbgbuf, pMsg->msgdata, pMsg->flags & CAN_DLC );
							KdPrint (( DRVNAME "msgdata %i: %s; %i\n", i, dbgbuf, dbgbuf[5] ));
							
							pMsg++; pUciMsg++; msgcnt++; buffer[0]++;
						}

						/*KdPrint (( DRVNAME "WRITE: 5\n" ));*/
						UsbBuildInterruptOrBulkTransferRequest(
							&urb, 
							sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
							devExt->BulkOUTPipe, 
							buffer, 
							NULL, 
							(sizeof(UCI_CANMSG)*(i+1))+1, /* 64 */
							USBD_TRANSFER_DIRECTION_OUT | USBD_SHORT_TRANSFER_OK, 
							NULL
							);

						/*KdPrint (( DRVNAME "WRITE: 6\n" ));*/

						ntStatus 
							= Tc_CallToUSBExt ( deviceObject, &urb, FALSE );

						KdPrint (( DRVNAME "WRITE: returned ntStatus(%X), UrbStatus(%X)\n", ntStatus, urb.UrbHeader.Status ));

						/*
						 * obolete: || devExt->WriteThread || devExt->IoQueueState == InActive
						 * because introduced a timeout for wait for single object to avoid dead lock
						 */
						if ( !NT_SUCCESS(ntStatus) || Irp->Cancel )
						{
							Irp->IoStatus.Information
								= ( msgcnt - i ) * sizeof(CANMSG);
							KdPrint (( DRVNAME "WRITE: FAILED to write bulk out endpoint; ntStatus(%X); UsbStatus(%X); msgcnt(%i); i(%i); transfered(%i)\n", ntStatus, urb.UrbHeader.Status, msgcnt, i, Irp->IoStatus.Information ));
							
							ntStatus = STATUS_UNSUCCESSFUL;	
							__leave;
						}

						/*
						 * ensure that these instructions are executed at PASSIVE_LEVEL, 
						 * otherwise someone will wait a very long long time...
						 * it need to be lower than DISPATCH_LEVEL because of the timeout
						 */
						ntStatus 
							= KeWaitForSingleObject ( &devExt->WriteIntEvent, Executive, KernelMode, FALSE, &devExt->WriteTimeout );
						/*KdPrint (( DRVNAME "WRITE: 7\n" ));*/
						if ( ntStatus != STATUS_SUCCESS || Irp->Cancel )
						{
							Irp->IoStatus.Information
								= ( msgcnt - i ) * sizeof(CANMSG);
							KdPrint (( DRVNAME "WRITE: FAILED; timeout while waiting for write interrupt; ntStatus(%X); msgcnt(%i); i(%i); transfered(%i)\n", ntStatus, msgcnt, i, Irp->IoStatus.Information ));
							/* 
								this situation is very bad dude :(
								seems the firmware hangs
								to avoid critical situations, we stop working now
							*/
							Tc_StopCanDL ( devExt );
							__leave;
						}

						transfered
							= ( (devExt->WriteIntBuffer.st_code >> 4) & 0x0f );
						if ( transfered != buffer[0] )
						{
							Irp->IoStatus.Information
								= ( msgcnt - i + transfered ) * sizeof(CANMSG);
							KdPrint (( DRVNAME "WRITE: FAILED; transferred (%i/%i) less than submitted to bus; ntStatus(%X); msgcnt(%i); i(%i); transfered(%i)\n", transfered, buffer[0], ntStatus, msgcnt, i, Irp->IoStatus.Information ));
							ntStatus = STATUS_UNSUCCESSFUL;					
							__leave;
						}
					}
					Irp->IoStatus.Information = mcount * sizeof(CANMSG);
				}
				__finally
				{
					Tc_FinishIrp0 ( Irp, ntStatus );
					IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
				}
			}
		}

		KeAcquireInStackQueuedSpinLock ( &devExt->WriteLock, &hLockLock );
		/*----------------------------------------------------------------*/
			devExt->WritePending = FALSE;
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLockLock );

		KeSetEvent ( &devExt->WriteStopEvent, 0, FALSE );

		KdPrint (( DRVNAME "WRITE: ---+++ StartWriteIo end +++-----!\n" ));
    }
}

