/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Common routines for tcan_usb driver
 *
 * TODO:
 *
 * History of changes:
 *	 2005/03/15 - 0.1 - first version
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "tcan_usb.h"
#include "ioctls.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_CallToUSB)
#pragma alloc_text( PAGE, Tc_ConjureDescriptorStringToBuffer)
#endif 

/******************************************************************************
 
   Tc_CallToUSB

 Description:
	Make a call to the usb subsystem by means of an internal device io control
	request.

 Execution IRQL:
	PASSIVE_LEVEL

  @deviceObject - the device object
  @urb

 returns a ntStatus

******************************************************************************/
NTSTATUS Tc_CallToUSBExt
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PURB			urb,
 IN BOOLEAN			wait
)
{
	PIRP				Irp;
	KEVENT				usbEvent;
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	IO_STATUS_BLOCK		ioStatus;
	PIO_STACK_LOCATION	ioStack;

	KdPrint (( DRVNAME "Tc_CallToUSB\n" ));
	PAGED_CODE ();
	ASSERT ( KeGetCurrentIrql() == PASSIVE_LEVEL );	

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	KeInitializeEvent ( &usbEvent, NotificationEvent, FALSE );


	Irp = IoBuildDeviceIoControlRequest (
		IOCTL_INTERNAL_USB_SUBMIT_URB,
		devExt->NextDeviceObject, 
		NULL, 
		0, 
		NULL, 
		0, 
		TRUE, 
		&usbEvent, 
		&ioStatus);

	if ( !Irp )
	{
		KdPrint((DRVNAME "insufficient resources, failed to allocate irp for urb\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	ioStack
		= IoGetNextIrpStackLocation ( Irp );

	ioStack->Parameters.Others.Argument1
		= (PVOID) urb;

	ntStatus
		= IoCallDriver ( devExt->NextDeviceObject, Irp );

	if ( ntStatus == STATUS_PENDING )
	{
		if ( wait )
			KeWaitForSingleObject ( &usbEvent, Executive, KernelMode, FALSE, &devExt->UsbTimeout );
		else
			KeWaitForSingleObject ( &usbEvent, Executive, KernelMode, FALSE, NULL );
		ntStatus = ioStatus.Status;
	}

	KdPrint (( DRVNAME "Tc_CallToUSB ntStatus(%X)\n", ntStatus ));
	return ntStatus;
}

NTSTATUS Tc_CallToUSB
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PURB			urb
)
{
	return Tc_CallToUSBExt ( deviceObject, urb, FALSE );
}

/******************************************************************************
 
   Tc_ConjureDescriptorStringToBuffer

 Description:
	Read in a string descriptor and write it in the output buffer of the irp

  @deviceObject - the device object
  @index
  @Irp
  @outLength

 returns a ntStatus

******************************************************************************/
NTSTATUS Tc_ConjureDescriptorStringToBuffer
(
 PDEVICE_OBJECT	deviceObject,
 UCHAR			index,
 PIRP			Irp,
 LONG			outLength
)
{
	URB						Urb;
	LONG					size;
	PULONG					oBuffer;
	NTSTATUS				ntStatus;
	UCHAR					Buffer[0xFF];
	PDEVICE_EXTENSION		devExt;
	PUSB_STRING_DESCRIPTOR	sd;

	PAGED_CODE ();
	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	UsbBuildGetDescriptorRequest (
		&Urb,
		sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
		USB_STRING_DESCRIPTOR_TYPE,
		index,
		devExt->DefaultLanguage,
		Buffer,
		NULL,
		sizeof(Buffer),
		NULL );

	ntStatus 
		= Tc_CallToUSB ( deviceObject, &Urb );

	if ( !NT_SUCCESS(ntStatus) )
		return ntStatus;

	sd = (PUSB_STRING_DESCRIPTOR) Buffer;

	size = sd->bLength-2;

	if ( outLength < size )
		return STATUS_BUFFER_TOO_SMALL;

	oBuffer = (PULONG) Irp->AssociatedIrp.SystemBuffer;
	RtlZeroMemory( oBuffer, outLength);
	RtlCopyMemory( oBuffer, sd->bString, size );
	Irp->IoStatus.Information = size;

	return STATUS_SUCCESS;
}


/******************************************************************************
 
   Tc_AcquireLock

 Description:

  @pQueue
  @Irql

 returns a ntStatus

******************************************************************************/
VOID Tc_AcquireLock
(
 IN  PIO_CSQ	pQueue,
 OUT PKIRQL		Irql
)
{
	PDEVICE_EXTENSION devExt;
    
	devExt
		= CONTAINING_RECORD(pQueue, DEVICE_EXTENSION, IoQueue);

    KeAcquireSpinLock ( &devExt->IoQueueLock, Irql );
}


/******************************************************************************
 
   Tc_ReleaseLock

 Description:

  @pQueue
  @Irql

 no return value

******************************************************************************/
VOID Tc_ReleaseLock
(
 IN  PIO_CSQ	pQueue,
 OUT KIRQL		Irql
)
{
	PDEVICE_EXTENSION devExt;

    devExt
		= CONTAINING_RECORD(pQueue, DEVICE_EXTENSION, IoQueue);
    
    KeReleaseSpinLock(&devExt->IoQueueLock, Irql);
}


/******************************************************************************
 
   Tc_InsertIrp

 Description:

  @pQueue
  @Irp

 no return value

******************************************************************************/
VOID Tc_InsertIrp
(
 IN PIO_CSQ pQueue,
 IN PIRP	Irp
)
{
    PDEVICE_EXTENSION devExt;

	devExt
		= CONTAINING_RECORD ( pQueue, DEVICE_EXTENSION, IoQueue );

    InsertTailList( &devExt->IoQueueHead, &Irp->Tail.Overlay.ListEntry );
	InterlockedIncrement ( &devExt->ReadCount );
}


/******************************************************************************
 
   Tc_PeekNextIrp

 Description:

  @pQueue
  @Irp
  @PeekContext

 returns a pointer to an IRP

******************************************************************************/
PIRP Tc_PeekNextIrp
(
 IN PIO_CSQ			pQueue,
 IN PIRP			Irp,
 IN PVOID	PeekContext
)
{
	PIRP				nextIrp;
    PLIST_ENTRY			next;
	PLIST_ENTRY			IoQueueHead;
	PDEVICE_EXTENSION	devExt;

    devExt
		= CONTAINING_RECORD ( pQueue, DEVICE_EXTENSION, IoQueue );
    
	IoQueueHead = &devExt->IoQueueHead;

    if( Irp == NULL )       
        next = IoQueueHead->Flink;
    else
        next = Irp->Tail.Overlay.ListEntry.Flink;

    nextIrp = NULL;

    while ( next != IoQueueHead ) 
	{
		PQUEUECONTEXT		pc;
		PIO_STACK_LOCATION	ioStack;
        nextIrp 
			= CONTAINING_RECORD ( next, IRP, Tail.Overlay.ListEntry );

        ioStack = IoGetCurrentIrpStackLocation(nextIrp);
        
		/*
		  NULL - context does not matter
		  TRUE - Write
		  FALSE - Read
		*/
		
		pc = (PQUEUECONTEXT) PeekContext;

		if ( pc->Context == &devExt->Read )
		{
			KdPrint (( DRVNAME "PeekContext Read\n" ));
			if ( ioStack->MajorFunction == IRP_MJ_READ )
				return nextIrp;
		}
		else
		{
			if ( pc->Context == &devExt->Write )
			{
				KdPrint (( DRVNAME "PeekContext Write\n" ));
				if ( ioStack->MajorFunction == IRP_MJ_WRITE )
					return nextIrp;
			}
			else
			{
				if ( pc->Context == &devExt->Interrupt )
				{
					KdPrint (( DRVNAME "PeekContext Interrupt\n" ));
					if ( ioStack->MajorFunction == IRP_MJ_DEVICE_CONTROL )
						/* && ioStack->Parameters.DeviceIoControl.IoControlCode == IOCTL_GET_INTERRUPT) */
						return nextIrp;
				}
				else
				{
					if ( pc->Context == &devExt->NoMatter )
					{
						KdPrint (( DRVNAME "PeekContext NoMatter\n" ));
						if ( pc->FileObject == NULL )
							return nextIrp;

						if ( pc->FileObject == ioStack->FileObject )
							return nextIrp;					
					}
				}
			}
		}          
        next = next->Flink;
    }
	return NULL;    
}


/******************************************************************************
 
   Tc_RemoveIrp

 Description:

  @pQueue
  @Irp

 no return value

******************************************************************************/
VOID Tc_RemoveIrp
(
 IN PIO_CSQ	pQueue,
 IN PIRP	Irp
)
{
	PDEVICE_EXTENSION devExt;

	/*
	   Makro to get the device extension from the queue object
	*/
    devExt
		= CONTAINING_RECORD(pQueue, DEVICE_EXTENSION, IoQueue);
    
    RemoveEntryList ( &Irp->Tail.Overlay.ListEntry );
	InterlockedDecrement ( &devExt->ReadCount );
}


/******************************************************************************
 
   Tc_CompleteCanceledIrp

 Description:

  @pQueue
  @Irp

 no return value

******************************************************************************/
VOID Tc_CompleteCanceledIrp
(
 IN PIO_CSQ	pQueue,
 IN PIRP	Irp
)
{
    KdPrint (( DRVNAME "queued irp cancelled\n" ));
	Tc_FinishIrp2 ( Irp, STATUS_CANCELLED );
}


/******************************************************************************
 
   Tc_FinishIrp

 Description:
	Helper routine to complete an Irp, with various conditions,
	noted by a number

  returns a ntStatus

******************************************************************************/
NTSTATUS Tc_FinishIrp
(
 IN PIRP		Irp, 
 IN NTSTATUS	ntStatus, 
 IN ULONG_PTR	Info
)
{
	Irp->IoStatus.Status = ntStatus;
	Irp->IoStatus.Information = Info;
	IoCompleteRequest ( Irp, IO_NO_INCREMENT );
	return ntStatus;
}

/******************************************************************************/
NTSTATUS Tc_FinishIrp2
(
 IN PIRP		Irp, 
 IN NTSTATUS	ntStatus
)
{
	return Tc_FinishIrp ( Irp, ntStatus, 0 );
}

/******************************************************************************/
NTSTATUS Tc_FinishIrp1
(
 IN PIRP		Irp
)
{
	return Tc_FinishIrp ( Irp, STATUS_SUCCESS, 0 );
}

/******************************************************************************/
NTSTATUS Tc_FinishIrp0
(
 IN PIRP		Irp, 
 IN NTSTATUS	ntStatus
)
{
	Irp->IoStatus.Status = ntStatus;
	IoCompleteRequest ( Irp, IO_NO_INCREMENT );
	return ntStatus;
}