/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Interrupt routines for tcan_usb driver
 *
 * TODO:
 *		Change to Spin Lock to lock the boolean flag instead of the buffer
 *			Must be tested first. Neccessary, cause of the IoMode was introduced after the lock was build in
 *
 * History of changes:
 *	 2005/03/15 - 0.1 - first version
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "tcan_usb.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_DeleteInterruptResources)
#endif 

/******************************************************************************
 
   Tc_DeleteInterruptResources

 Description: 
	free memory ressources of urb and irp allocated by StartDevice	

  @devExt - the device extension

 no return value

******************************************************************************/
VOID Tc_DeleteInterruptResources
(
 IN PDEVICE_EXTENSION devExt
)
{
	PAGED_CODE ();
	if ( devExt->IntUrb )
	{
		ExFreePool ( devExt->IntUrb );
		devExt->IntUrb = NULL;
	}

	if ( devExt->IntIrp )
	{
		IoFreeIrp ( devExt->IntIrp );
		devExt->IntIrp = NULL;
	}
}


/******************************************************************************
 
   Tc_OnInterrupt

 Description:
	This is the completion routine for a read operation on the interrupt pipe.

 Note:
	Because of the execution at DISPATCH_LEVEL, next releases should disable some
	of the spin locks
	Workaround: the can devices firmware behaves some strange with buffer overruns
				thus we disconnect from it if more than 10 overruns comes sequently

  @deviceObject - the device Extension
  @Irp
  @devExt

 returns a ntStatus value

******************************************************************************/
NTSTATUS Tc_OnInterrupt
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp, 
 PDEVICE_EXTENSION	devExt
)
{
	KLOCK_QUEUE_HANDLE hLock;

	KdPrint (( DRVNAME "INTPOLL: Tc_OnInterrupt...level(%i)\n", KeGetCurrentIrql() ));

	KeAcquireInStackQueuedSpinLock ( &devExt->InterruptLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->IntPollPending = FALSE;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	if ( NT_SUCCESS(Irp->IoStatus.Status) && !Irp->Cancel  && devExt->CanStarted )
	{
		PUCI_INT_DATA IntBuffer;
		KdPrint (( DRVNAME "INTPOLL: ---- Interrupt! -----\n" ));

		IntBuffer 
			= &devExt->IntBuffer;

		/* 
		   we need the spin lock, to prevent something from dereference the event, while we are signaling it
		   or the next interrupt event change the values
		 */
		if ( (devExt->IntEvent && !devExt->IntSet) || !devExt->IoMode )
		{
			KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLock );
			/*----------------------------------------------------------------*/
				devExt->IntEventBuffer = *IntBuffer;
			/*----------------------------------------------------------------*/
			KeReleaseInStackQueuedSpinLock ( &hLock );
			devExt->IntSet = TRUE;
			if ( devExt->IoMode )
			{
				KeSetEvent ( devExt->IntEvent, EVENT_INCREMENT, FALSE );
			}
			else
			{
				KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) FALSE );
			}
		}

		if ( IntBuffer->st_code & UCI_ST_FW_RXFIFO_OVR )
		{
			InterlockedIncrement ( &devExt->numFWOvr );
			if ( InterlockedIncrement ( &devExt->cacheFWOvr ) >= UCI_FAIL_BEHAVE_COUNT )
			{
				Tc_StopCanDL ( devExt );
				/*PSTOP_WORKITEM wi = (PSTOP_WORKITEM) ExAllocatePool ( NonPagedPool, sizeof(STOP_WORKITEM));
				if ( !wi )
				{
					Tc_StopCanDL ( deviceObject );
				}
				else
				{
					wi->wi = IoAllocateWorkItem ( deviceObject );
					if ( !wi->wi )
					{
						ExFreePool ( wi );
						Tc_StopCanDL ( deviceObject );
					}
					else
					{
						wi->deviceObject = deviceObject;
						IoQueueWorkItem ( wi->wi, (PIO_WORKITEM_ROUTINE) WAStop_Device, CriticalWorkQueue, wi );
					}
				}*/
			}
			KdPrint (( DRVNAME "INTPOLL: Firmware fifo overruns count(%i)\n", devExt->numFWOvr ));
		}
		/*else InterlockedExchange ( &devExt->cacheFWOvr, 0 );*/
	
		KdPrint (( DRVNAME "INTPOLL: st_code: %X\n", IntBuffer->st_code ));
		#if DBG
		if ( IntBuffer->errors )		KdPrint (( DRVNAME "errors: %X\n", IntBuffer->errors ));
		if ( IntBuffer->overruns )		KdPrint (( DRVNAME "overruns: %X\n", IntBuffer->overruns ));
		if ( IntBuffer->irq )			KdPrint (( DRVNAME "irq: %X\n", IntBuffer->irq ));
		if ( IntBuffer->status )		KdPrint (( DRVNAME "status: %X\n", IntBuffer->status));
		if ( IntBuffer->err_code_cap )	KdPrint (( DRVNAME "err_code_cap: %X\n", IntBuffer->err_code_cap ));
		if ( IntBuffer->err_warn_lim )	KdPrint (( DRVNAME "err_warn_lim: %X\n", IntBuffer->err_warn_lim ));
		if ( IntBuffer->rx_err_cnt )	KdPrint (( DRVNAME "rx_err_cnt: %X\n", IntBuffer->rx_err_cnt ));
		if ( IntBuffer->tx_err_cnt )	KdPrint (( DRVNAME "tx_err_cnt: %X\n", IntBuffer->tx_err_cnt ));
		if ( IntBuffer->fw_fifo_ovr )	KdPrint (( DRVNAME "fw_fifo_ovr: %X\n", IntBuffer->fw_fifo_ovr ));
		if ( IntBuffer->mode )			KdPrint (( DRVNAME "mode: %X\n", IntBuffer->mode ));
		#endif
		if ( devExt->WritePending )
		{
			devExt->WriteIntBuffer = *IntBuffer;
			if ( KeSetEvent ( &devExt->WriteIntEvent, EVENT_INCREMENT, FALSE ) )
			{
				TRUE;
				#if DBG
				KdPrint (( DRVNAME "INTPOLL: Interrupt indicates write flow mismatch; lasting event available\n" ));
				#endif
			}
		}
		if ( devExt->cacheFWOvr < UCI_FAIL_BEHAVE_COUNT )
			Tc_StartPolling ( devExt, TRUE );
	}	
	else
	{
		KdPrint (( DRVNAME "INTPOLL: IRP canceled; ntStatus(%X); UrbStatus(%X)\n",
			Irp, Irp->IoStatus.Status, URB_STATUS(devExt->IntUrb) ));
	}

	IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );

	return STATUS_MORE_PROCESSING_REQUIRED;
}

/******************************************************************************
 
   WAStop_Device

 Description:
  WorkItem-Routine to stop the can device.
	
 Execution IRQL:
	PASSIVE_LEVEL

  @wi - the work item context

 no return value

******************************************************************************/
VOID WAStop_Device
(
 IN PSTOP_WORKITEM wi
)
{
	PAGED_CODE ();
	Tc_StopCan ( wi->deviceObject, FALSE );
	IoFreeWorkItem ( wi->wi );
	ExFreePool ( wi );
}

/******************************************************************************
 
   StartInterruptIo

 Description:
  This routine processes the io-queue for all interrupt irps. If there's no 
  irp in the queue, it ends.
  Irp comes either from DispatchReadWrite or StartIoDpc.
	
 Execution IRQL:
	DISPATCH_LEVEL

  @deviceObject - the device object
  @Irp

 no return value

******************************************************************************/
VOID Tc_StartInterruptIo
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	PDEVICE_EXTENSION	devExt;
	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	KdPrint (( DRVNAME "INT: ---+++ Tc_StartInterruptIo start at level (%X) +++-----!\n", KeGetCurrentIrql() ));

	/*
	   if the can device is deactivated meanwhile we fail the irp and sends an stop event
	*/
	if ( !devExt->CanStarted )
		Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
	else
	{
		KLOCK_QUEUE_HANDLE	hLock;
		NTSTATUS			ntStatus;
		ntStatus = STATUS_UNSUCCESSFUL;

		Irp->IoStatus.Information = 0;
		
		ntStatus
			= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );
		
		if ( !NT_SUCCESS(ntStatus) )
			Tc_FinishIrp2 ( Irp, STATUS_UNSUCCESSFUL );
		else
		{
			__try
			{
				KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLock );
				/*----------------------------------------------------------------*/
					*((PUCI_INT_DATA) Irp->AssociatedIrp.SystemBuffer) = devExt->IntEventBuffer;
				/*----------------------------------------------------------------*/
				KeReleaseInStackQueuedSpinLock ( &hLock );

				ntStatus = STATUS_SUCCESS;

				Irp->IoStatus.Information = sizeof(UCI_INT_DATA);

				/* Reset Interrupt flag to allow new Interrupts */
				devExt->IntSet = FALSE;
			}
			__finally
			{
				Tc_FinishIrp0 ( Irp, ntStatus );
				IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			}
		}
	}
	KdPrint (( DRVNAME "INT: ---+++ Tc_StartInterruptIo end +++-----!\n" ));

	/* KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) FALSE ); */
}