/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Dispatch routine for the IOCTL_Xxx requests for tcan_usb driver
 *
 * TODO:
 *
 * History of changes:
 *	 2005/03/15 - 0.1 - first version
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


#include "tcan_usb.h"
#include "ioctls.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_InitAndStartCan)
#pragma alloc_text( PAGE, Tc_StopCan)
#endif 

/******************************************************************************
 
   Tc_DispatchControl

 Description:
	This routine handles the DeviceIoControl requests	

  @deviceObject - the device object
  @Irp

 returns a ntStatus

******************************************************************************/
NTSTATUS Tc_DispatchControl
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
)
{
	ULONG				retInfo;
	ULONG				inLength;
	ULONG				outLength;
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	ioStack;
	
	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	ntStatus
		= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );

	if ( !NT_SUCCESS(ntStatus) )
		return Tc_FinishIrp ( Irp, ntStatus, 0 );
	
	Irp->IoStatus.Information = 0;
	retInfo = 0;

	ioStack
		= IoGetCurrentIrpStackLocation ( Irp );
	
	inLength
		= ioStack->Parameters.DeviceIoControl.InputBufferLength;
	
	outLength
		= ioStack->Parameters.DeviceIoControl.OutputBufferLength;

	switch ( ioStack->Parameters.DeviceIoControl.IoControlCode )
	{
		case IOCTL_GET_INTERRUPT:
		{
			KLOCK_QUEUE_HANDLE hLock;

			KdPrint (( DRVNAME " IOCTL_GET_INTERRUPT\n" ));

			if ( outLength < sizeof(UCI_INT_DATA) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				break;
			}

			if ( devExt->IoMode )
			{
				KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLock );
					*((PUCI_INT_DATA) Irp->AssociatedIrp.SystemBuffer) = devExt->IntEventBuffer;
				KeReleaseInStackQueuedSpinLock ( &hLock );
				devExt->IntSet = FALSE;
				ntStatus = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(UCI_INT_DATA);
			}
			else
			{
				if ( devExt->IoQueueState != InActive )
				{
					IoCsqInsertIrp ( &devExt->IoQueue, Irp, NULL );
					IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
					KdPrint (( DRVNAME "queued\n" ));
					if ( devExt->IntSet )
						KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) FALSE );
					return STATUS_PENDING;
				}
				else
				{
					ntStatus = STATUS_DEVICE_NOT_READY;
					Irp->IoStatus.Information = 0;
				}
			}

			break;
		}

		case IOCTL_SET_READ_EVENT:
		{
			HANDLE				hEvent;
			PKEVENT				pEvent;
			KLOCK_QUEUE_HANDLE	hLock;

			KdPrint (( DRVNAME " IOCTL_SET_READ_EVENT\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				break;
			}

			#if DBG
			if ( devExt->ReadEvent != NULL )
				KdPrint (( DRVNAME "INFO; an event is already registered, set new one\n" ));
			#endif

			if ( inLength < sizeof(HANDLE) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; ( size mismatch ) object given is not an event handle\n" ));
				break;
			}

			hEvent 
				= *( (PHANDLE) Irp->AssociatedIrp.SystemBuffer );
			if ( !hEvent )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED to set read event; object given is not a valid event handle\n" ));
				break;
			}

			ntStatus = ObReferenceObjectByHandle (
				hEvent,
				EVENT_MODIFY_STATE,
				*ExEventObjectType,
				Irp->RequestorMode,
				(PVOID*) &pEvent,
				NULL);
			if ( !NT_SUCCESS(ntStatus) )
			{
				KdPrint (( DRVNAME "FAILED look up for an event handle; ntStatus(%X)\n", ntStatus ));
				break;
			}

			KeAcquireInStackQueuedSpinLock ( &devExt->ReadEventLock, &hLock );
			if ( devExt->ReadEvent )
				ObDereferenceObject ( devExt->ReadEvent );
			devExt->ReadEvent = pEvent;
			KeReleaseInStackQueuedSpinLock ( &hLock );
			
			devExt->ReadFileObject
				= ioStack->FileObject;

			break;
		}

		case IOCTL_SET_INTERRUPT_EVENT:
		{
			HANDLE				hEvent;
			PKEVENT				pEvent;
			KLOCK_QUEUE_HANDLE	hLock;

			KdPrint (( DRVNAME " IOCTL_SET_INTERRUPT_EVENT\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				break;
			}

			#if DBG
			if ( devExt->IntEvent != NULL )
				KdPrint (( DRVNAME "INFO; an event is already registered\n" ));
			#endif

			if ( inLength < sizeof(HANDLE) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; ( size mismatch ) object given is not an event handle\n" ));
				break;
			}

			hEvent
				= *( (PHANDLE) Irp->AssociatedIrp.SystemBuffer );
			if (!hEvent)
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED to set interrupt event; object given is not a valid event handle\n" ));
				break;
			}

			ntStatus = ObReferenceObjectByHandle (
				hEvent,
				EVENT_MODIFY_STATE,
				*ExEventObjectType,
				Irp->RequestorMode,
				(PVOID*) &pEvent,
				NULL);
			if ( !NT_SUCCESS(ntStatus) )
			{
				KdPrint (( DRVNAME "FAILED look up for an event handle; ntStatus(%X)\n", ntStatus ));
				break;
			}

			KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLock );
			if ( devExt->IntEvent )
				ObDereferenceObject ( devExt->IntEvent );
			devExt->IntEvent = pEvent;
			KeReleaseInStackQueuedSpinLock ( &hLock );	
			
			devExt->IntFileObject
				= ioStack->FileObject;

			break;
		}

		case IOCTL_SET_IO_MODE:
		{
			CIOMODE iomode;

			KdPrint (( DRVNAME " IOCTL_SET_IO_MODE\n" ));

			if ( devExt->CanStarted )
			{
				ntStatus = STATUS_DEVICE_BUSY;
				KdPrint (( DRVNAME "FAILED to set io mode; can device is busy\n" ));
				break;
			}

			if ( inLength < sizeof(CIOMODE) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; ( size mismatch ) mode given is not an io mode\n" ));
				break;
			}

			iomode
				= *( (PCIOMODE) Irp->AssociatedIrp.SystemBuffer );
			if (!iomode)
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED to set io mode; data given is not a valid io mode\n" ));
				break;
			}

			devExt->IoMode = iomode;
			ntStatus = STATUS_SUCCESS;

			break;
		}

		case IOCTL_SET_RX_BUFFER_SIZE:
		{
			ULONG size;
			KdPrint (( DRVNAME " IOCTL_SET_RX_BUFFER_SIZE\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			if ( devExt->CanStarted )
			{
				ntStatus = STATUS_DEVICE_BUSY;
				KdPrint (( DRVNAME "FAILED; can device must be stopped before set buffer size\n" ));
				break;
			}

			if ( inLength > sizeof(ULONG) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			size
				= *( (PULONG) Irp->AssociatedIrp.SystemBuffer );
			if ( size < 0x100 )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			devExt->RXFifoSize = size;
			devExt->RXFifo.head = 0;
			devExt->RXFifo.tail = 0;
			devExt->RXFifo.status = CANMSG_BUF_EMPTY;
			
			ntStatus = STATUS_SUCCESS;

			break;
		}

		case IOCTL_START:
		{
			URB					urb;
			UCHAR				ubuffer[5];
			KLOCK_QUEUE_HANDLE	hLock;

			KdPrint (( DRVNAME " IOCTL_START\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			if ( outLength < sizeof(unsigned char) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			/*-------------------------------------------------------------------------
			 *
			 *  submit the can device a start request
			 *
			 *-------------------------------------------------------------------------*/
			/*
			   urb.bmRequestType = 0xC0; (0x1100000) direction dev to host, vendor req type to dev, reservedbits = 0
			 */
			
			if ( devExt->CanStarted )
			{
				KdPrint (( DRVNAME "FAILED: Can device already running\n" ));
				ntStatus = STATUS_UNSUCCESSFUL;
				break;
			}

			/* First we need to start the polling for interrupt and reads */
			ntStatus = Tc_InitAndStartCan ( devExt, (PUCHAR) &ubuffer );
			if ( !NT_SUCCESS(ntStatus) )
			{
				KdPrint (( DRVNAME "FAILED to start pollings; ntStatus(%X)\n", ntStatus ));
				break;
			}
			else
			{
				*((PULONG) Irp->AssociatedIrp.SystemBuffer) = ubuffer[0];
				Irp->IoStatus.Information = 1;

				if ( ubuffer[0] & CAN_RESET_MODE_REQUEST )
				{
					KdPrint (( DRVNAME "Can Start returns %.2x\n", ubuffer[0] ));
					ntStatus = STATUS_UNSUCCESSFUL;
					break;
				}
				KeAcquireInStackQueuedSpinLock ( &devExt->CanStartedLock, &hLock );
				/*----------------------------------------------------------------*/
					devExt->CanStarted = TRUE;
				/*----------------------------------------------------------------*/
				KeReleaseInStackQueuedSpinLock ( &hLock );
			}
			break;
		}

		case IOCTL_STOP:
		{
			KdPrint (( DRVNAME " IOCTL_STOP\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}

			/*-------------------------------------------------------------------------
			 *
			 *  submit the can device a stop request
			 *
			 *-------------------------------------------------------------------------*/
			/*
			   urb.bmRequestType = 0x40; (0x01000000) direction host to dev, vendor req type to device, reservedbits = 0
			*/

			if ( !devExt->CanStarted || devExt->DevState != Working)
			{
				KdPrint (( DRVNAME "FAILED: Can device already stopped\n" ));
				ntStatus = STATUS_UNSUCCESSFUL;
				break;
			}

			ntStatus = Tc_StopCan ( deviceObject , TRUE );
			break;
		}

		case IOCTL_RESET:
		{
			URB urb;
			KdPrint (( DRVNAME " IOCTL_RESET\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}

			/*-------------------------------------------------------------------------
			 *
			 *  submit the can device a reset request
			 *
			 *-------------------------------------------------------------------------*/
			/*
			   urb.bmRequestType = 0x40; (0x01000000) direction host to dev, vendor req type to device, reservedbits = 0
			*/
			if ( devExt->CanStarted )
			{
				ntStatus = Tc_StopCan ( deviceObject , TRUE );
				if ( !NT_SUCCESS(ntStatus) )
				{
					KdPrint (( DRVNAME "FAILED to start pollings; ntStatus(%X)\n", ntStatus ));
					break;
				}
			}

			/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
			UsbBuildVendorRequest (
				&urb,
				URB_FUNCTION_VENDOR_DEVICE,						/* 0x0017 */
				sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
				0,
				0,
				UCI_CAN_CHIP_ACCESS,		/* 0xB0 */
				0, 
				UCI_CAN_RESET,				/* 0x07 */
				NULL,
				NULL,
				0,
				NULL );

			ntStatus 
				= Tc_CallToUSB ( deviceObject, &urb );

			if ( !NT_SUCCESS(ntStatus) )
				KdPrint((DRVNAME "FAILED to reset can device; ntStatus(%X); UrbStatus(%X)\n", ntStatus, urb.UrbHeader.Status));
			else
				Irp->IoStatus.Information = 0;
			
			/*
			   maybe start the can device afterwards, but it's the usual way to start it manually
			*/
			/*
			ntStatus = InitAndStartCan ( devExt );
			if (!NT_SUCCESS(ntStatus))
			{
				KdPrint (( DRVNAME "FAILED to start pollings; ntStatus(%X)\n", ntStatus ));
				break;
			}
			*/

			break;
		}

		case IOCTL_DRIVER_VERSION:
		{
			KdPrint (( DRVNAME " IOCTL_DRIVER_VERSION\n" ));
			/*-------------------------------------------------------------------------
			 *
			 *  return the internal driver version
			 *
			 *-------------------------------------------------------------------------*/
			if ( outLength < sizeof (ULONG) )
				ntStatus = STATUS_BUFFER_TOO_SMALL;
			else
			{			
				*( (PULONG) Irp->AssociatedIrp.SystemBuffer ) = DRVERSION;
				Irp->IoStatus.Information = sizeof(DRVERSION);
				ntStatus = STATUS_SUCCESS;
			}
			break;
		}

		case IOCTL_GET_VID:
		{
			KdPrint (( DRVNAME " IOCTL_GET_VID\n" ));
			if ( outLength < sizeof(SHORT) )
				ntStatus = STATUS_BUFFER_TOO_SMALL;
			else
			{			
				*( (PSHORT) Irp->AssociatedIrp.SystemBuffer ) = devExt->DescDevice.idVendor;
				Irp->IoStatus.Information = sizeof( devExt->DescDevice.idVendor );
				ntStatus = STATUS_SUCCESS;
			}
			break;
		}

		case IOCTL_GET_PID:
		{
			KdPrint (( DRVNAME " IOCTL_GET_PID\n" ));
			if ( outLength < sizeof(SHORT) )
				ntStatus = STATUS_BUFFER_TOO_SMALL;
			else
			{			
				*( (PSHORT) Irp->AssociatedIrp.SystemBuffer ) = devExt->DescDevice.idProduct;
				Irp->IoStatus.Information = sizeof ( devExt->DescDevice.idProduct );
				ntStatus = STATUS_SUCCESS;
			}
			break;
		}

		case IOCTL_GET_REVISION:
		{
			KdPrint (( DRVNAME " IOCTL_GET_REVISION\n" ));
			if ( outLength < sizeof(SHORT) )
				ntStatus = STATUS_BUFFER_TOO_SMALL;
			else
			{			
				*( (PSHORT) Irp->AssociatedIrp.SystemBuffer ) = devExt->DescDevice.bcdDevice;
				Irp->IoStatus.Information = sizeof ( devExt->DescDevice.bcdUSB );
				ntStatus = STATUS_SUCCESS;
			}
			break;
		}

		case IOCTL_GET_MANUFACTURER:
		{
			KdPrint (( DRVNAME " IOCTL_GET_MANUFACTURER\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			ntStatus
				= Tc_ConjureDescriptorStringToBuffer ( deviceObject, devExt->DescDevice.iManufacturer, Irp, outLength );
			break;
		}

		case IOCTL_GET_PRODUCT:
		{
			KdPrint (( DRVNAME " IOCTL_GET_PRODUCT\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			ntStatus 
				= Tc_ConjureDescriptorStringToBuffer ( deviceObject, devExt->DescDevice.iProduct, Irp, outLength );
			break;
		}

		case IOCTL_GET_SERIALNUMBER:
		{
			KdPrint (( DRVNAME " IOCTL_GET_SERIALNUMBER\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			ntStatus
				= Tc_ConjureDescriptorStringToBuffer ( deviceObject, devExt->DescDevice.iSerialNumber, Irp, outLength );
			break;
		}

		case IOCTL_SET_BAUD:
		{
			URB		urb;
			USHORT	baud;

			KdPrint (( DRVNAME " IOCTL_SET_BAUD\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}

			if ( inLength != sizeof(CANBAUD) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			/*-------------------------------------------------------------------------
			 *
			 *  set the accept code
			 *
			 *-------------------------------------------------------------------------*/
			baud = (USHORT) *((PULONG) Irp->AssociatedIrp.SystemBuffer);
			if ( !baud || baud >= CAN_BAUD_ENDVALUE )
			{
				ntStatus = STATUS_INVALID_DEVICE_REQUEST;
				break;
			}

			KdPrint (( DRVNAME "Tc_DispatchControl() - IOCTL_TCAN_SET_BAUD baud: %X\n", baud ));

			/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
			UsbBuildVendorRequest (
				&urb, 
				URB_FUNCTION_VENDOR_DEVICE, 
				sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), 
				0,										/* .TransferFlags */
				0, 
				UCI_CAN_CHIP_ACCESS,					/* bRequest 0xB0 */
				0,										/* .Value */
				( (baud << 8) | UCI_SET_BUS_TIMINGS ),	/* .Index baud.0x04 */
				NULL,									/* .TransferBuffer */
				NULL, 
				0, 
				NULL );

			ntStatus 
				= Tc_CallToUSB ( deviceObject, &urb );

			if ( !NT_SUCCESS(ntStatus) )
				KdPrint (( DRVNAME "Error %X (USBD status code %X) trying to set baudrate\n", ntStatus, urb.UrbHeader.Status ));
			else
				Irp->IoStatus.Information = 0;
			break;
		}

		case IOCTL_SET_ACC_CODE:
		{
			URB		urb;
			CHAR	buffer[5];

			KdPrint (( DRVNAME " IOCTL_SET_ACC_CODE\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			if ( inLength != sizeof(ULONG) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			RtlZeroMemory ( buffer, 5 );

			/*-------------------------------------------------------------------------
			 *
			 *  set the accept code
			 *
			 *-------------------------------------------------------------------------*/
			RtlCopyMemory ( (void *) buffer, (CHAR *) Irp->AssociatedIrp.SystemBuffer, 4 );
			/*memcpy(buffer, (CHAR *) Irp->AssociatedIrp.SystemBuffer, 4); */

			KdPrint (( DRVNAME " acc code in buffer: %.2x, %.2x, %.2x, %.2x\n", buffer[0], buffer[1], buffer[2], buffer[3] ));

			/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
			UsbBuildVendorRequest ( 
				&urb, 
				URB_FUNCTION_VENDOR_DEVICE, 
				sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), 
				0,						/* .TransferFlags */
				0, 
				UCI_CAN_CHIP_ACCESS,	/* bRequest 0xB0 */
				0x00,					/* .Value */
				UCI_SET_ACCEPT_CODE,	/* .Index 0x05 */
				buffer,					/* .TransferBuffer */
				NULL, 
				4, 
				NULL );

			ntStatus 
				= Tc_CallToUSB ( deviceObject, &urb );

			if ( !NT_SUCCESS(ntStatus) )
				KdPrint (( DRVNAME "FAILED to set acc code; ntStatus(%X); UsbStatus(%X)\n", ntStatus, urb.UrbHeader.Status ));
			else
				Irp->IoStatus.Information = 0;
			break;
		}

		case IOCTL_GET_ACC_CODE:
		{
			URB		urb;
			USHORT	i;
			PULONG	poBuffer;
			CHAR	buffer[5];

			RtlZeroMemory ( buffer, 5 );

			KdPrint (( DRVNAME " IOCTL_GET_ACC_CODE\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			if ( outLength != sizeof(ULONG) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			poBuffer = (PULONG) Irp->AssociatedIrp.SystemBuffer;

			/*---------------------------------------------------------------------------------------------
			 *
			 *  read the can register for accept code
			 *
			 *		In order to read the acceptance code, we need to read each register with an urb
			 *		in each urb we use the values 16, 17, 18, 19 for ACCEPTANCE_CODE_0 - ACCEPTANCE_CODE_3
			 *
			 *		bmRequestType - 1100000, dev to host, vendor request type
			 *---------------------------------------------------------------------------------------------*/
			for (i = 0; i < 4; i++)
			{
				/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
				UsbBuildVendorRequest ( 
					&urb, 
					URB_FUNCTION_VENDOR_DEVICE, 
					sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), 
					USBD_TRANSFER_DIRECTION_IN, /* .TransferFlags */
					0,							/* .RequestTypeReservedBits */
					UCI_CAN_CHIP_ACCESS,		/* bRequest 0xB0 */
					16+i,						/* .Value */
					UCI_READ_CAN_REGISTER,		/* .Index */
					buffer+i,					/* .TransferBuffer */
					NULL, 
					1,							/* .TransferBufferLength */
					NULL );

				ntStatus
					= Tc_CallToUSB ( deviceObject, &urb );

				if (!NT_SUCCESS(ntStatus))
				{
					KdPrint (( DRVNAME "FAILED to get acc code register %i ntStatus(%X), UsbStatus(%X)\n", i, ntStatus, urb.UrbHeader.Status ));
					break;
				}
			}

			KdPrint (( DRVNAME "get acceptance code returns %.2x %.2x %.2x %.2x\n", buffer[0], buffer[1], buffer[2], buffer[3] ));
			RtlCopyMemory ( poBuffer, buffer, 4 );
			/* memcpy ( poBuffer, buffer, 4 ); */
			Irp->IoStatus.Information = 4;

			break;
		}

		case IOCTL_SET_ACC_MASK:
		{
			URB		urb;
			CHAR	buffer[5];

			RtlZeroMemory ( buffer, 5 );

			KdPrint (( DRVNAME " IOCTL_SET_ACC_MASK\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			if ( inLength != sizeof(ULONG) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			/*-------------------------------------------------------------------------
			 *
			 *  set the mask for accept
			 *
			 *-------------------------------------------------------------------------*/
			RtlCopyMemory ( buffer, (CHAR *) Irp->AssociatedIrp.SystemBuffer, 4 );
			/* memcpy ( buffer, (CHAR *) Irp->AssociatedIrp.SystemBuffer, 4 ); */

			KdPrint (( DRVNAME "mask in buffer: %02x, %02x, %02x, %02x\n", buffer[0], buffer[1], buffer[2], buffer[3] ));

			/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
			UsbBuildVendorRequest (
				&urb, 
				URB_FUNCTION_VENDOR_DEVICE, 
				sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), 
				0,							/* .TransferFlags */
				0,							/* .RequestTypeReservedBits */
				UCI_CAN_CHIP_ACCESS,		/* bRequest 0xB0 */
				0,							/* .Value */
				UCI_SET_ACCEPT_MASK,		/* .Index */
				buffer,						/* .TransferBuffer */
				NULL, 
				4,							/* .TransferBufferLength */
				NULL );

			ntStatus
				= Tc_CallToUSB ( deviceObject, &urb );

			if ( !NT_SUCCESS(ntStatus) )
				KdPrint (( DRVNAME "FAILED to set acc mask; ntStatus(%X); UsbStatus(%X)\n", ntStatus, urb.UrbHeader.Status ));
			else
				Irp->IoStatus.Information = 0;
			break;
		}

		case IOCTL_GET_ACC_MASK:
		{
			URB		urb;
			USHORT	i;
			PULONG	poBuffer;
			CHAR	buffer[5];

			RtlZeroMemory ( buffer, 5 );

			KdPrint (( DRVNAME " IOCTL_GET_ACC_MASK\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			if ( outLength != sizeof(ULONG) )
			{
				ntStatus = STATUS_INVALID_PARAMETER;
				KdPrint (( DRVNAME "FAILED; invalid buffer size given\n" ));
				break;
			}

			poBuffer = (PULONG) Irp->AssociatedIrp.SystemBuffer;

			/*---------------------------------------------------------------------------------------------
			 *
			 *  read the can register for accept mask
			 *
			 *		In order to read the acceptance mask, we need to read each register with an urb
			 *		in each urb we use the values 20, 21, 22, 23 for ACCEPTANCE_MASK_0 - ACCEPTANCE_MASK_3
			 *
			 *		bmRequestType - 1100000, dev to host, vendor request type
			 *---------------------------------------------------------------------------------------------*/
			for ( i = 0; i < 4; i++ )
			{
				/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
				UsbBuildVendorRequest ( 
					&urb, 
					URB_FUNCTION_VENDOR_DEVICE, 
					sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), 
					USBD_TRANSFER_DIRECTION_IN, /* .TransferFlags */
					0,							/* .RequestTypeReservedBits */
					UCI_CAN_CHIP_ACCESS,		/* bRequest 0xB0 */
					20+i,						/* .Value */
					UCI_READ_CAN_REGISTER,		/* .Index */
					buffer+i,					/* .TransferBuffer */
					NULL, 
					1,							/* .TransferBufferLength */
					NULL );

				ntStatus
					= Tc_CallToUSB ( deviceObject, &urb );

				if ( !NT_SUCCESS(ntStatus ))
				{
					KdPrint (( DRVNAME "FAILED to get acc mask register %i ntStatus(%X), UsbStatus(%X)\n", i, ntStatus, urb.UrbHeader.Status ));
					break;
				}
			}

			KdPrint (( DRVNAME "get acceptance mask returns %.2x %.2x %.2x %.2x\n", buffer[0], buffer[1], buffer[2], buffer[3] ));
			RtlCopyMemory ( poBuffer, buffer, 4 );
			/* memcpy ( poBuffer, buffer, 4 ); */
			Irp->IoStatus.Information = 4;

			break;
		}

		/*  deprecated
		case IOCTL_GET_INTERRUPTCOUNT:
		{
			if ( outLength < 4 )
				ntStatus = STATUS_NO_MEMORY;
			else
			{	
				KdPrint (( DRVNAME "current interrupt count %X\n", devExt->numInts ));		
				*((PULONG) Irp->AssociatedIrp.SystemBuffer) = devExt->numInts;
				ntStatus = STATUS_SUCCESS;
				Irp->IoStatus.Information = 4;
			}
			break;
		}
		*/

	   /*---------------------------------------------------------------------------------------------
		*
		*   NOT INTENDED FOR PRODUCT USE !!!
		*
		*---------------------------------------------------------------------------------------------*/
		case IOCTL_GET_AVR_STATUS:
		{
			URB		urb;
			UCHAR	ubuffer[5];

			KdPrint (( DRVNAME " IOCTL_GET_AVR_STATUS - not intended for product use !!!\n" ));

			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			
			RtlZeroMemory ( ubuffer, 5 );
			/*
			   check whether avr is there and ready ( use my request 0xB1 and rqtype IN )
			  
				ret = usb_control_msg (
				avrp_devs[dev_num].udev,	// unsigned int pipe
				 VENDOR_RQ_IN,			// requesttype (0xC0); 0x11000000, dev to host, vendor, to device
				 RQ_GET_AVR_STATUS,		// (0xB1) request; Vendor specific requests of avrpipe firmware
				 0,				// __u16 value,
				 0,				// __u16 index,
				 buf,			// void * data,
				 4,				// __u16 size, 
				 100			// int timeout
				 );
			*/
			/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
			UsbBuildVendorRequest ( 
				&urb,											/* urb */
				URB_FUNCTION_VENDOR_INTERFACE,					/* UrbHeader.Function */
				sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),	/* UrbHeader.Length */
				USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK, /* UrbControlVendorClassRequest.TransferFlags */
				0,										/* .RequestTypeReservedBits */
				RQ_GET_AVR_STATUS,						/* .Request (0xB1) */
				0,										/* .Value */
				0,										/* .Index */
				ubuffer,									/* .TransferBuffer */
				NULL,									/* .TransferBufferMDL */
				4,										/* .TransferBufferLength */
				NULL									/* .UrbLink */
				);

			ntStatus
				= Tc_CallToUSB ( deviceObject, &urb );

			if ( !NT_SUCCESS(ntStatus) )
			{
				KdPrint (( DRVNAME "FAILED to get avr status by control message, urbStatus(%X)\n", urb.UrbHeader.Status ));
				break;
			}
			KdPrint (( DRVNAME "get avr status returned %X\n", ubuffer[0] ));
			break;
		}

		case IOCTL_MTNC_STALLQUEUE:
		{
			KdPrint (( DRVNAME " IOCTL_MTNC_STALLQUEUE\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			/*-------------------------------------------------------------------------
			 *
			 *  Maintenance ioctl
			 *
			 *-------------------------------------------------------------------------*/
			Tc_StallQueueAwait ( devExt );
			ntStatus = STATUS_SUCCESS;
			break;
		}

		case IOCTL_MTNC_RESTARTQUEUE:
		{
			KdPrint (( DRVNAME " IOCTL_MTNC_RESTARTQUEUE\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			/*-------------------------------------------------------------------------
			 *
			 *  Maintenance ioctl
			 *
			 *-------------------------------------------------------------------------*/
			Tc_RestartQueue ( deviceObject );
			ntStatus = STATUS_SUCCESS;
			break;
		}

		case IOCTL_MTNC_ABORTPIPE:
		{
			URB		urb;
			CHAR	pipe;
			USBD_PIPE_HANDLE ph;
			ph = NULL;

			KdPrint (( DRVNAME " IOCTL_MTNC_ABORTBULKOUT\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			/*-------------------------------------------------------------------------
			 *
			 *  Maintenance ioctl
			 *
			 *-------------------------------------------------------------------------*/
			pipe = *((CHAR *)Irp->AssociatedIrp.SystemBuffer);
			switch ( pipe )
			{
				case 1:
                    ph = devExt->InterruptINPipe;
					KdPrint (( DRVNAME "ABORTPIPE: Interrupt %i\n", pipe ));
					break;
				case 2:
					ph = devExt->BulkINPipe;
					KdPrint (( DRVNAME "ABORTPIPE: Bulk IN %i\n", pipe ));
					break;
				case 3:
					ph = devExt->BulkOUTPipe;
					KdPrint (( DRVNAME "ABORTPIPE: BULK Out %i\n", pipe ));
					break;
				default:
					KdPrint (( DRVNAME "ABORTPIPE: FAILED; invalid pipe; %i\n", pipe ));
					break;
			}
			if ( ph )
			{
				/*
					abort the pipe
				*/	
				urb.UrbHeader.Length
					= (USHORT) sizeof(struct _URB_PIPE_REQUEST);
				urb.UrbHeader.Function
					= URB_FUNCTION_ABORT_PIPE;
				urb.UrbPipeRequest.PipeHandle
					= ph;

				ntStatus = Tc_CallToUSB ( deviceObject, &urb );
			}
			if ( !NT_SUCCESS(ntStatus) )
			{
				TRUE;
				KdPrint (( DRVNAME "ABORTPIPE: FAILED; abort the pipe; UrbStatus(%X)\n", urb.UrbHeader.Status ));
			}
			break;
		}

		case IOCTL_MTNC_RESETDEVICE:
		{
			PIRP	Irp;
			KEVENT	sig;
			IO_STATUS_BLOCK ioStatus;

			KdPrint (( DRVNAME " IOCTL_MTNC_RESETDEVICE\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			/*-------------------------------------------------------------------------
			 *
			 *  Maintenance ioctl
			 *
			 *-------------------------------------------------------------------------*/
			/*
				reset the device
			*/			
			KeInitializeEvent ( 
				&sig, 
				NotificationEvent, 
				FALSE );

			Irp = IoBuildDeviceIoControlRequest (
				IOCTL_INTERNAL_USB_RESET_PORT,
				devExt->NextDeviceObject,
				NULL, 0, NULL, 0, TRUE,
				&sig,
				&ioStatus );
			if ( !Irp )
			{
				ntStatus = STATUS_INSUFFICIENT_RESOURCES;
				KdPrint (( DRVNAME "RESETBULKOUT: FAILED; low resources; ntStatus(%X)\n", ntStatus ));
				break;
			}

			ntStatus = IoCallDriver ( devExt->NextDeviceObject, Irp );
			if ( ntStatus == STATUS_PENDING )
			{
				KeWaitForSingleObject (
					&sig, 
					Executive, 
					KernelMode, 
					FALSE, 
					NULL );
				ntStatus = ioStatus.Status;
			}
			
			if ( !NT_SUCCESS(ntStatus) )
			{
				TRUE;
				KdPrint (( DRVNAME "RESETBULKOUT: FAILED; ntStatus(%X)\n", ntStatus ));
			}
			break;
		}

		case IOCTL_MTNC_RESETPIPE:
		{
			URB		urb;
			CHAR	pipe;
			USBD_PIPE_HANDLE ph;
			ph = NULL;

			KdPrint (( DRVNAME " IOCTL_MTNC_RESETBULKOUT\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			/*-------------------------------------------------------------------------
			 *
			 *  Maintenance ioctl
			 *
			 *-------------------------------------------------------------------------*/
			pipe = *((CHAR *)Irp->AssociatedIrp.SystemBuffer);
			switch ( pipe )
			{
				case 1:
                    ph = devExt->InterruptINPipe;
					KdPrint (( DRVNAME "RESETPIPE: Interrupt %i\n", pipe ));
					break;
				case 2:
					ph = devExt->BulkINPipe;
					KdPrint (( DRVNAME "RESETPIPE: Bulk IN %i\n", pipe ));
					break;
				case 3:
					ph = devExt->BulkOUTPipe;
					KdPrint (( DRVNAME "RESETPIPE: BULK Out %i\n", pipe ));
					break;
				default:
					KdPrint (( DRVNAME "RESETPIPE: FAILED; invalid pipe; %i\n", pipe ));
					break;
			}
			if ( ph )
			{
				/*
					reset the pipe
				*/	
				urb.UrbHeader.Length
					= (USHORT) sizeof(struct _URB_PIPE_REQUEST);
				urb.UrbHeader.Function
					= URB_FUNCTION_RESET_PIPE;
				urb.UrbPipeRequest.PipeHandle
					= ph;

				ntStatus = Tc_CallToUSB ( deviceObject, &urb );
			}
			if ( !NT_SUCCESS(ntStatus) )
			{
				TRUE;
				KdPrint (( DRVNAME "RESETPIPE: FAILED; reset the pipe; UrbStatus(%X)\n", urb.UrbHeader.Status ));
			}
			break;
		}

		case IOCTL_MTNC_DECONFIGURE:
		{
			URB		urb;

			KdPrint (( DRVNAME " IOCTL_MTNC_DECONFIGURE\n" ));
			if ( devExt->DevState != Working )
			{
				ntStatus = STATUS_DEVICE_NOT_READY;
				KdPrint (( DRVNAME "FAILED; device pnp state not working\n" ));
				break;
			}
			/*-------------------------------------------------------------------------
			 *
			 *  Maintenance ioctl
			 *
			 *-------------------------------------------------------------------------*/
			UsbBuildSelectConfigurationRequest (
				&urb, 
				sizeof(struct _URB_SELECT_CONFIGURATION), 
				NULL );

			ntStatus = Tc_CallToUSB ( deviceObject, &urb );
			if ( !NT_SUCCESS(ntStatus) )
			{
				TRUE;
				KdPrint (( DRVNAME "DECONFIGURE: Failed to deconfigure the device; UrbStatus(%X)\n", urb.UrbHeader.Status ));
			}
			break;
		}

		default:
		ntStatus = STATUS_INVALID_DEVICE_REQUEST;
		break;
	}

	IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
	return Tc_FinishIrp0 ( Irp, ntStatus );
}


/******************************************************************************
 
   Tc_InitAndStartCan

 Description:
	Initialize RX Fifo buffer, allocate memory for and set status
	of operations to initial values
	as well as start polling of interrupt and bulk endpoints

  Note: This routine has to be called only at PASSIVE_LEVEL, because of the
		apc for the event in CallToUSB.
		Maybe it's better for the latency and priority if passing down the
		Irp and receycle it at DISPATCH_LEVEL

  @devExt - the device Extension

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_InitAndStartCan
(
 IN PDEVICE_EXTENSION	devExt,
 IN PUCHAR				ubuffer
)
{
	URB urb;
	NTSTATUS ntStatus;
	HANDLE hThread;
	KLOCK_QUEUE_HANDLE hLock;

	KdPrint (( DRVNAME "Tc_InitAndStartCan()\n" ));
	PAGED_CODE ();
	
	ntStatus = STATUS_UNSUCCESSFUL;
	/*
	   Initialize Read Buffer
	 */
	RtlZeroMemory ( &devExt->RXFifo, sizeof(BULK_FIFO) );

	devExt->RXFifo.head = 0;
	devExt->RXFifo.tail = 0;
	devExt->RXFifo.status = CANMSG_BUF_EMPTY;
	devExt->RXFifo.msg = NULL;

	devExt->ReadCount = 0;
	devExt->RXFifoOverrun = 0;
	devExt->numFWOvr = 0;
	devExt->cacheFWOvr = 0;

	devExt->WritePending = FALSE;
	devExt->ReadPending = FALSE;

	__try
	{
		devExt->RXFifo.msg
			= (PCANMSG) ExAllocatePool ( NonPagedPool, devExt->RXFifoSize * sizeof(CANMSG) );

		if ( !devExt->RXFifo.msg )
		{
			KdPrint (( DRVNAME "STARTCAN: Failed to allocate fifo-buffer\n" ));
			__leave;
		}

	   /*-------------------------------------------------------------------------
		*
		*  Create thread resources for write operations
		*
		*-------------------------------------------------------------------------*/
		KeInitializeSemaphore ( &devExt->WriteThreadSemaphore, 0, MAXLONG );
		devExt->WriteThreadStop = FALSE;
		ntStatus = PsCreateSystemThread (
			&hThread,
			0,
			NULL,
			0,
			NULL,
			Tc_StartWriteIoThread,
			devExt->DeviceObject );

		if( !NT_SUCCESS(ntStatus) )
		{
			KdPrint (( DRVNAME "STARTCAN: Failed to create systemthread\n" ));
			return ntStatus;
		}

		ObReferenceObjectByHandle (
			hThread,
			THREAD_ALL_ACCESS,
			NULL,
			KernelMode,
			(PVOID*) &devExt->WriteThread,
			NULL );

		ZwClose ( hThread );

		KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
		/*----------------------------------------------------------------*/
			devExt->IoQueueState = Active;
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );

	   /*-------------------------------------------------------------------------
		*
		*  Begin polling the endpoints
		*
		*-------------------------------------------------------------------------*/
		ntStatus = Tc_StartPolling ( devExt, FALSE );
		ntStatus = ( ntStatus == STATUS_PENDING ? STATUS_SUCCESS : ntStatus );
		if (!NT_SUCCESS(ntStatus))
		{
			KdPrint (( DRVNAME "STARTCAN: Failed to start read poll\n" ));
			__leave;
		}

		ntStatus = Tc_StartPolling ( devExt, TRUE );
		ntStatus = ( ntStatus == STATUS_PENDING ? STATUS_SUCCESS : ntStatus );
		if ( !NT_SUCCESS(ntStatus) )
		{
			KdPrint (( DRVNAME "STARTCAN: Failed to start int poll\n" ));
			__leave;
		}
			
		/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
		UsbBuildVendorRequest (
			&urb, 
			URB_FUNCTION_VENDOR_DEVICE,						/* 0x0017 */
			sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), 
			USBD_TRANSFER_DIRECTION_IN,						/* (1<<0) */
			0, 
			UCI_CAN_CHIP_ACCESS, /* 0xB0 */
			0, 
			UCI_CAN_START,		/* 0x08 */
			ubuffer, 
			NULL, 
			1, 
			NULL );

		ntStatus 
			= Tc_CallToUSB ( devExt->DeviceObject, &urb );

		if ( !NT_SUCCESS(ntStatus) )
		{
			KdPrint (( DRVNAME "FAILED to start can device; ntStatus(%X); UrbStatus(%X)\n", ntStatus, urb.UrbHeader.Status ));
			__leave;
		}
	}
	__finally
	{
		if ( !NT_SUCCESS(ntStatus) )
		{
			KLOCK_QUEUE_HANDLE hLock1;
			KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock1 );
			/*----------------------------------------------------------------*/
				devExt->IoQueueState = InActive;
			/*----------------------------------------------------------------*/
			KeReleaseInStackQueuedSpinLock ( &hLock1 );

			if ( devExt->RXFifo.msg )
				ExFreePool ( devExt->RXFifo.msg );
			devExt->RXFifo.msg = NULL;

			if ( devExt->WriteThread )
			{
				devExt->WriteThreadStop = TRUE;

				KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, FALSE );

				KeWaitForSingleObject (
					&devExt->WriteThread,
					Executive,
					KernelMode,
					FALSE,
					NULL );

				ObDereferenceObject ( &devExt->WriteThread );
				devExt->WriteThread = NULL;
			}

			if ( devExt->ReadPollIrp )
				Tc_StopPolling ( devExt, FALSE );

			if ( devExt->IntIrp )
				Tc_StopPolling ( devExt, TRUE );
		}
	}
	return ntStatus;
}


/******************************************************************************
	
	  Tc_StopCan

 Description:
	Stop the Can device and free the memory ressources
	
  @deviceObject - the device Extension
  @CanControl

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_StopCan
(
 IN PDEVICE_OBJECT	deviceObject,
 IN BOOLEAN			CanControl
)
{
	NTSTATUS ntStatus;
	KLOCK_QUEUE_HANDLE hLock;
	PDEVICE_EXTENSION devExt;

	KdPrint (( DRVNAME "Tc_StopCan() - %s\n", ( CanControl ? "with USB-Stop" : "without USB-Stop" ) ));
	PAGED_CODE ();

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	ntStatus = STATUS_SUCCESS;

	if ( CanControl )
	{
		/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
		URB urb;
		UsbBuildVendorRequest(
			&urb,
			URB_FUNCTION_VENDOR_DEVICE,						/* 0x0017 */
			sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
			0,
			0, 
			UCI_CAN_CHIP_ACCESS,		/* 0xB0 */
			0, 
			UCI_CAN_STOP,				/* 0x09 */
			NULL,
			NULL,
			0,
			NULL
			);

		ntStatus
			= Tc_CallToUSB ( deviceObject, &urb );

		#if DBG
		if ( !NT_SUCCESS(ntStatus) )
			KdPrint (( DRVNAME "FAILED to stop can device; ntStatus(%X); UrbStatus(%X)\n", ntStatus, urb.UrbHeader.Status ));
		#endif
	}
	
	/* 1: should possible to be called multiple */
	Tc_StopQueueAwait ( devExt );

	if ( devExt->WriteThread )
	{
		/*devExt->WriteThreadStop = TRUE;

		KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, TRUE );

		KeWaitForSingleObject (
			&devExt->WriteThread,
			Executive,
			KernelMode,
			FALSE,
			NULL );
		*/

		ObDereferenceObject ( &devExt->WriteThread );
		devExt->WriteThread = NULL;
	}

	Tc_StopPolling ( devExt, TRUE );

	Tc_StopPolling ( devExt, FALSE );

	if ( devExt->RXFifo.msg != NULL )
	{
		ExFreePool( devExt->RXFifo.msg );
		devExt->RXFifo.msg = NULL;
	}

	/* 2: Cancel all Irps in IoQueue */
	Tc_ClearIoQueue ( devExt );

	KeAcquireInStackQueuedSpinLock ( &devExt->CanStartedLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->CanStarted = FALSE;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	return ntStatus;
}


/******************************************************************************
	
	  Tc_StopCanDL

 Description:
	Stop the Can device and free the memory ressources for execution at dispatch level
	
  @deviceObject - the device Extension

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_StopCanDL
(
 IN PDEVICE_EXTENSION	devExt
)
{
	NTSTATUS ntStatus;
	KLOCK_QUEUE_HANDLE hLock;

	KdPrint (( DRVNAME "Tc_StopCanDL()\n" ));

	ntStatus = STATUS_SUCCESS;
	
	/*KdPrint (( DRVNAME "CRITSTOP: 0.1\n" ));*/
	KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
	/*----------------------------------------------------------------*/
		/*KdPrint (( DRVNAME "CRITSTOP: 0.2\n" ));*/
		devExt->IoQueueState = InActive;
		/*KdPrint (( DRVNAME "CRITSTOP: 0.3\n" ));*/
		devExt->WriteThreadStop = TRUE;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	/*KdPrint (( DRVNAME "CRITSTOP: 1\n" ));*/
	/*
	   Why say FALSE for the wait parameter, and why it crashes when wait equals TRUE and
	   is executed within the spin lock above
	   The kernel use a dispatcher database lock ( w. oney ) for it's thread scheduling
	   things. ReleaseSemaphore, SetEvent, WaitEvent and so on needs to aquire this lock.
	   this means it raise the irql above dispatch_level, don't lower it and the thread's terminating
	   above passive_level at all. Results in a bug check worker_thread_....
    */
	
	KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, FALSE );

	/*KdPrint (( DRVNAME "CRITSTOP: 2\n" ));*/
	if ( devExt->WriteThread )
	{
		ObDereferenceObject ( &devExt->WriteThread );
		devExt->WriteThread = NULL;
	}

	/*KdPrint (( DRVNAME "CRITSTOP: 3\n" ));*/
	Tc_StopPolling ( devExt, TRUE );

	/*KdPrint (( DRVNAME "CRITSTOP: 4\n" ));*/
	Tc_StopPolling ( devExt, FALSE );

	/*KdPrint (( DRVNAME "CRITSTOP: 5\n" ));*/
	if ( devExt->RXFifo.msg != NULL )
	{
		ExFreePool( devExt->RXFifo.msg );
		devExt->RXFifo.msg = NULL;
	}

	/*KdPrint (( DRVNAME "CRITSTOP: 6\n" ));*/
	/* 2: Cancel all Irps in IoQueue */
	Tc_ClearIoQueue ( devExt );

	/*KdPrint (( DRVNAME "CRITSTOP: 7\n" ));*/
	KeAcquireInStackQueuedSpinLock ( &devExt->CanStartedLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->CanStarted = FALSE;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	if ( devExt->IntEvent )
	{
		KLOCK_QUEUE_HANDLE hLock;
		KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLock );
		/*----------------------------------------------------------------*/							
			RtlZeroMemory ( &devExt->IntEventBuffer, sizeof(UCI_INT_DATA) );
			devExt->IntEventBuffer.st_code = 0xa0;
			devExt->IntSet = TRUE;
			if ( devExt->IoMode )
			{
				if (devExt->IntEvent)
					KeSetEvent(devExt->IntEvent, EVENT_INCREMENT, FALSE);
			}
			else
			{
				/* Start a dpc for processing an interrupt Irp
				   check IRQ-Level
				   --> we can also do this without a dpc
			     */
				KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) FALSE );
			}
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );
	}
	devExt->PhysicalReconnectNeeded = TRUE;

	Tc_SetPnpState ( devExt, Stopped );
	
	devExt->PhysicalReconnectNeeded = TRUE;
	IoInvalidateDeviceState ( devExt->Pdo );
	/*IoInvalidateDeviceRelations ( devExt->Pdo, BusRelations );*/

	KdPrint (( DRVNAME "CRITSTOP: done\n" ));
	return ntStatus;
}