/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Power Management Routines - very early alpha state - prototype
 *
 * TODO:
 *		Testing
 *
 * History of changes:
 *	 2005/03/23 - 0.1 - first version
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "tcan_usb.h"

NTSTATUS Tc_HandleQueryPowerSystem
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);

NTSTATUS Tc_HandleQueryPowerDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);

NTSTATUS Tc_HandleSetPowerSystem
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);

NTSTATUS Tc_HandleSetPowerDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);

NTSTATUS Tc_CreateDevicePowerIrpCR
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp,
 IN PVOID			Context
);

NTSTATUS Tc_SetPowerDeviceCR
(
 IN PDEVICE_OBJECT	_deviceObject,
 IN PIRP			Irp,
 IN PVOID			Context
);

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_DispatchPower)
#endif 

/******************************************************************************
 
   Tc_DispatchPower

 Description:
	Dispatcher routine for power management irps

  @deviceObject - the device object
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_DispatchPower
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	ULONG				minfc;
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	ioStack;

	KdPrint (( DRVNAME "Tc_DispatchPower()\n" ));
	PAGED_CODE ();
	
    ioStack
		= IoGetCurrentIrpStackLocation ( Irp );

    devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

    if( devExt->DevState == Removed ) 
	{
        PoStartNextPowerIrp ( Irp );
		return Tc_FinishIrp ( Irp, STATUS_DELETE_PENDING, 0 );
    }

    if ( devExt->DevState == NeverStarted ) 
	{
        PoStartNextPowerIrp ( Irp );
        IoSkipCurrentIrpStackLocation ( Irp );
        return PoCallDriver ( devExt->NextDeviceObject, Irp );
    }

    minfc = ioStack->MinorFunction;
    switch ( minfc ) 
	{    
		case IRP_MN_SET_POWER:
			switch ( ioStack->Parameters.Power.Type )
			{

				case SystemPowerState:
					KdPrint (( DRVNAME "PWR - IRP_MN_SET_POWER - System\n" ));
					ntStatus = Tc_HandleSetPowerSystem ( devExt, Irp );
					break;

				case DevicePowerState:
					KdPrint (( DRVNAME "PWR - IRP_MN_SET_POWER - Device\n" ));
					ntStatus = Tc_HandleSetPowerDevice ( devExt, Irp );
					break;
			}
			break;

		case IRP_MN_QUERY_POWER:    
			switch ( ioStack->Parameters.Power.Type )
			{
				case SystemPowerState:
					KdPrint (( DRVNAME "PWR - IRP_MN_QUERY_POWER - System\n" ));
					ntStatus = Tc_HandleQueryPowerSystem ( devExt, Irp );
					break;

				case DevicePowerState:
					KdPrint (( DRVNAME "PWR - IRP_MN_QUERY_POWER - Device\n" ));
					ntStatus = Tc_HandleQueryPowerDevice ( devExt, Irp );
					break;
			}
			break;

		default:
			KdPrint (( DRVNAME "PWR - NOTOFINTEREST - %i\n", minfc ));
			PoStartNextPowerIrp ( Irp );
			IoSkipCurrentIrpStackLocation ( Irp );
			ntStatus
				= PoCallDriver ( devExt->NextDeviceObject, Irp );
			break;
    }
    return ntStatus;
}


/******************************************************************************
 
   Tc_HandleQueryPowerSystem

 Description:
	This IRP is send by the Power Manager, when the system wants to switch to
	the lower power level S1-S3.
	Depending on the WaitForCompletionOnSuspend Flag, we reject or wait for
	the completion of the irp

  @devExt - the device extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleQueryPowerSystem
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
	PIO_STACK_LOCATION ioStack;

	KdPrint (( DRVNAME "Tc_HandleQueryPowerSystem()\n" ));

    ioStack
		= IoGetCurrentIrpStackLocation ( Irp );

	/*
	   We stall the queue and don't wait
	*/
	devExt->IoQueueState = Stalled;

	PoStartNextPowerIrp ( Irp );
	IoSkipCurrentIrpStackLocation ( Irp );
	return PoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   Tc_HandleQueryPowerDevice

 Description:
	

  @devExt - the device extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleQueryPowerDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
	PIO_STACK_LOCATION ioStack;

    ioStack
		= IoGetCurrentIrpStackLocation ( Irp );

	KdPrint (( DRVNAME "Tc_HandleQueryPowerDevice() %X to %X\n",devExt->DevPower ,ioStack->Parameters.Power.State.SystemState ));

	devExt->IoQueueState = Stalled;

	PoStartNextPowerIrp ( Irp );
	IoSkipCurrentIrpStackLocation ( Irp );
	return PoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   Tc_HandleSetPowerSystem

 Description:
	A set system power state can only be send by the power manager to notify
	the driver of a state change. The state transistions can be:

	S0 to S0 - A query has failed
		- restart the queue in a completion routine

	SX to S0 - The system wake up maybe to change to a less power state
		- do nothing

	S0 to SX - A change to a less power state
		- set a completion routine and:
		  stop the queue, stop the can device, free all resources
		  this has to be done, because in one of these states, the power could
		  be lost, thus the firmware get lost and the device is not connected

  @devExt - the device extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleSetPowerSystem
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
	NTSTATUS			ntStatus;
	PIO_STACK_LOCATION	ioStack;
	SYSTEM_POWER_STATE	sysPowerState;

    ioStack
		= IoGetCurrentIrpStackLocation ( Irp );

	KdPrint (( DRVNAME "Tc_HandleSetPowerSystem() %X to %X\n",devExt->DevPower ,ioStack->Parameters.Power.State.SystemState ));

    sysPowerState
		= ioStack->Parameters.Power.State.SystemState;

	ntStatus = STATUS_SUCCESS;

	devExt->DevPowerNext =
		devExt->CanCapabilities.DeviceState[sysPowerState];

	if ( devExt->DevPowerNext == PowerDeviceD0 && devExt->DevPower == PowerDeviceD3 )
	{
		/*
		   don't create a device power irp, because the device will be removed surprise
		*/
		devExt->DevPower = devExt->DevPowerNext;
		PoStartNextPowerIrp ( Irp );
		IoSkipCurrentIrpStackLocation ( Irp );
	}
	else
	{
		IoCopyCurrentIrpStackLocationToNext ( Irp );
		IoSetCompletionRoutine (
			Irp,
			Tc_CreateDevicePowerIrpCR,
			devExt,
			TRUE,	/* invoke on success */
			TRUE,	/* error */
			TRUE );	/* and cancelation of irp */

		PoStartNextPowerIrp ( Irp );
	}

	return PoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   Tc_CreateDevicePowerIrpCR

 Description:
	

  @_deviceObject - the device object
  @Irp
  @Context - given by the originator

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_CreateDevicePowerIrpCR
(
 IN PDEVICE_OBJECT	_deviceObject,
 IN PIRP			Irp,
 IN PVOID			Context
)
{
	NTSTATUS			ntStatus;
	POWER_STATE			state;
	PDEVICE_EXTENSION	devExt;

	KdPrint (( DRVNAME "Tc_CreateDevicePowerIrpCR()\n" ));

	devExt
		= (PDEVICE_EXTENSION) Context;

	if (Irp->PendingReturned)
		IoMarkIrpPending ( Irp );

	state.DeviceState = devExt->DevPowerNext;
	
	ntStatus = PoRequestPowerIrp ( 
		devExt->Pdo,
		IRP_MN_SET_POWER,
		state,
		NULL,	/* completion routine */
		NULL, /* context to the completion routine */
		NULL );
	
	return Irp->IoStatus.Status = STATUS_SUCCESS;
}


/******************************************************************************
 
   Tc_HandleSetPowerDevice

 Description:
	A set device power state can be send by the power manager and the driver
	to change the power state. The state transistions can be:

	D0 to D0 - A query has failed
		- restart the queue in a completion routine

	DX to D0 - The system wake up maybe to change to a less power state
		- do nothing

	D0 to DX - A change to a less power state
		- set a completion routine and:
		  stop the queue, stop the can device, free all resources
		  this has to be done, because in one of these states, the power could
		  be lost, thus the firmware get lost and the device is not connected
	

  @devExt - the device extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleSetPowerDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
    NTSTATUS			ntStatus;
	DEVICE_POWER_STATE	state;
	PIO_STACK_LOCATION	ioStack;

    ioStack
		= IoGetCurrentIrpStackLocation ( Irp );

	KdPrint (( DRVNAME "Tc_HandleSetPowerDevice() %X to %X\n",devExt->DevPower ,ioStack->Parameters.Power.State.SystemState ));

    state
		= ioStack->Parameters.Power.State.DeviceState;

	/*
	   TODO: is there a possibility that state != devExt->DevPowerNext ??
	 			if yes, DevPowerNext should be corrected to state
	*/
    if ( state == PowerDeviceD0 && state < devExt->DevPower ) 
	{
        /*
           comming back to Working State either to go to a suspend state or to work
        
		   this shouldn't be neccessary, but who knows...
		 */ 
		devExt->IoQueueState = InActive;

        IoSkipCurrentIrpStackLocation ( Irp );
		PoStartNextPowerIrp ( Irp );

        ntStatus 
			= PoCallDriver ( devExt->NextDeviceObject, Irp );
	}
    else
	{
		IoCopyCurrentIrpStackLocationToNext ( Irp );

		IoSetCompletionRoutine (
				Irp, 
				(PIO_COMPLETION_ROUTINE) Tc_SetPowerDeviceCR,
				devExt, 
				TRUE, 
				TRUE, 
				TRUE );

		ntStatus
			= PoCallDriver ( devExt->NextDeviceObject, Irp );
    }

    return ntStatus;
}


/******************************************************************************
 
   Tc_SetPowerDeviceCR

 Description:
	

  @_deviceObject - the device object
  @Irp
  @Context - given by the originator

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_SetPowerDeviceCR
(
 IN PDEVICE_OBJECT	_deviceObject,
 IN PIRP			Irp,
 IN PVOID			Context
)
{
	POWER_STATE			state;
	PDEVICE_EXTENSION	devExt;

	devExt
		= (PDEVICE_EXTENSION) Context;

	KdPrint (( DRVNAME "Tc_SetPowerDeviceCR() %X to %X\n",devExt->DevPower ,devExt->DevPowerNext ));

	if ( Irp->PendingReturned )
		IoMarkIrpPending ( Irp );

	state.DeviceState = devExt->DevPowerNext;
	
	if ( devExt->DevPowerNext == PowerDeviceD0 && devExt->DevPowerNext == devExt->DevPower )
	{
		/*
		   a query has failed
			-> restart the stalled queue
		*/
		if ( devExt->IoQueueState == Stalled )
		{
			devExt->IoQueueState = Active;
			Tc_RestartQueue ( devExt->DeviceObject );
		}
	}
	else
	{
		if ( devExt->DevPowerNext > devExt->DevPower )
		{
			/*
			   change to suspend state
			  	--> it doesn't matter, which suspend state is intended to change to
			  		we must stop the queue and wait for ending of the current processing, 
			 		cancel all irps and stop the can device
			*/
			if ( devExt->CanStarted )
				Tc_StopCan ( devExt->DeviceObject, TRUE );
		}
	}
	
	devExt->DevPower = devExt->DevPowerNext;	
	return Irp->IoStatus.Status = STATUS_SUCCESS;
}