/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Driver entry and main routines for tcan_usb driver
 *
 * TODO:
 *
 * History of changes:
 *	 2005/05/29 - 0.2 - changed rules for pnp state. moved code between remove_device and stop_device
 *	 2005/03/15 - 0.1 - first version
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "tcan_usb.h"

NTSTATUS DriverEntry 
(
 IN PDRIVER_OBJECT	driverObject, 
 IN PUNICODE_STRING RegistryPath
);

#ifdef ALLOC_PRAGMA
#pragma alloc_text( INIT, DriverEntry )
#pragma alloc_text( PAGE, Tc_AddDevice)
#pragma alloc_text( PAGE, Tc_StartDevice)
#pragma alloc_text( PAGE, Tc_StopDevice)
#pragma alloc_text( PAGE, Tc_DriverUnload)
#pragma alloc_text( PAGE, Tc_RemoveDevice)
#pragma alloc_text( PAGE, Tc_DispatchCreate)
#pragma alloc_text( PAGE, Tc_DispatchCleanup)
#endif 

/******************************************************************************
 
   DriverEntry

 Description:
	Driver entry - see diploma thesis	

  @driverObject - the driver object
  @RegistryPath

 returns a ntstatus

******************************************************************************/
NTSTATUS DriverEntry 
(
 IN PDRIVER_OBJECT	driverObject, 
 IN PUNICODE_STRING RegistryPath
)
{
	KdPrint (( DRVNAME "DriverEntry USB-Tiny-CAN version %.8X Diploma Thesis, Chi Tai Dang (thesis at chi-tai dot info)\n", DRVERSION ));

	/*
		we need at least WDM version 1.10
		tcan_usb is tested only with Windows XP,
		so we ensure that WIN2K is at least the confronted kernel environment
		--------------------------------------------------------------------------------*/
	if ( !IoIsWdmVersionAvailable (0x01, 0x10) )
	{
		KdPrint (( DRVNAME "DriverEntry failed, wdm version 1.10 not detected\n" ));
		KdPrint (( DRVNAME "\tcurrently tcan_usb does not work with WIN9x!\n" ));
		return STATUS_UNSUCCESSFUL;
	}

	/*
		connect the DriverObject function pointers with the TinyCAN_ldr dispatch function
		--------------------------------------------------------------------------------*/
	driverObject->MajorFunction[IRP_MJ_CREATE] = Tc_DispatchCreate;
	driverObject->MajorFunction[IRP_MJ_CLOSE] = Tc_DispatchClose;	
	driverObject->MajorFunction[IRP_MJ_CLEANUP] = Tc_DispatchCleanup;	
	driverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = Tc_DispatchControl;
	driverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] = Tc_DispatchInternalControl;
	driverObject->MajorFunction[IRP_MJ_POWER] = Tc_DispatchPower;
	driverObject->MajorFunction[IRP_MJ_PNP] = Tc_DispatchPnp;
	driverObject->MajorFunction[IRP_MJ_READ] = Tc_DispatchReadWrite;
	driverObject->MajorFunction[IRP_MJ_WRITE] = Tc_DispatchReadWrite;

	/* to be tested ! to be enhanced ! */
	driverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = Tc_DispatchWmi;

	driverObject->DriverUnload = Tc_DriverUnload;
	driverObject->DriverExtension->AddDevice = Tc_AddDevice;

	return STATUS_SUCCESS;
}


/******************************************************************************
 
   Tc_AddDevice

 Description:

  @driverObject - the device Extension
  @PDO

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_AddDevice
(
 IN PDRIVER_OBJECT driverObject, 
 IN PDEVICE_OBJECT Pdo
)
{
	/*
	  decl
	*/
	WCHAR				buffer[32];
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	static LONG			deviceNumber;
	UNICODE_STRING		deviceName;
	PDEVICE_OBJECT		deviceObject;

	KdPrint (( DRVNAME "Tc_AddDevice\n" ));
	PAGED_CODE ();

	deviceNumber = -1;
	_snwprintf ( buffer, ( sizeof(buffer) / sizeof(buffer[0]) ), L"\\DosDevices\\CANUSB%d", InterlockedIncrement(&deviceNumber) );
	RtlInitUnicodeString ( &deviceName, buffer );

	ntStatus = IoCreateDevice (
		driverObject, 
		sizeof(DEVICE_EXTENSION), 
		&deviceName,
		FILE_DEVICE_UNKNOWN, 
		FILE_DEVICE_SECURE_OPEN, 
		FALSE,						/* This should be changed to TRUE for an exclusive device, one handle to be opened only */
		&deviceObject );

	if ( !NT_SUCCESS(ntStatus) )
	{
		KdPrint (( DRVNAME "FAILED IoCreateDevice; ntStatus(%X)\n", ntStatus ));
		return ntStatus;
	}
	KdPrint (( DRVNAME "Device created with name (%ws)\n", deviceName.Buffer ));

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;


	devExt->isFlashed = FALSE;
	devExt->DeviceObject = deviceObject;
	devExt->Pdo = Pdo;
	devExt->DevState = NeverStarted;
	devExt->DevStatePrev = NeverStarted;
	IoInitializeRemoveLock ( &devExt->RemoveLock, 0, 0, 0 );

	/*
	   Attach the device object to the device stack		
	*/
	devExt->NextDeviceObject
		= IoAttachDeviceToDeviceStack ( deviceObject, Pdo );

	if ( !devExt->NextDeviceObject )
	{						 
		KdPrint (( DRVNAME "FAILED IoAttachDeviceToDeviceStack\n" ));
		ntStatus = STATUS_DEVICE_REMOVED;
		IoDeleteDevice ( deviceObject );
		return ntStatus;
	}

	/*
	   Initialize Power Management
	*/
	deviceObject->Flags |= DO_POWER_PAGABLE;
	deviceObject->Flags |= DO_POWER_INRUSH;
	devExt->DevPower = PowerDeviceD0;
	devExt->DevPowerNext = PowerDeviceD0;

	deviceObject->Flags |= DO_DIRECT_IO;

	/*
	   Initialize the IoQueue
	*/
	KeInitializeSpinLock ( &devExt->IoQueueLock );
	KeInitializeSpinLock ( &devExt->IoQueueStateLock );
	InitializeListHead ( &devExt->IoQueueHead );

	IoCsqInitialize ( &devExt->IoQueue, Tc_InsertIrp, 
		Tc_RemoveIrp, Tc_PeekNextIrp, Tc_AcquireLock, 
		Tc_ReleaseLock, Tc_CompleteCanceledIrp );

	/*
	   Initialize IO-Queue DPC
	*/
	KeInitializeDpc ( &devExt->IoDpc, (PKDEFERRED_ROUTINE) Tc_StartIoDpc, deviceObject );

	/*
	   Initialize Write Thread
	*/
	KeInitializeSemaphore ( &devExt->WriteThreadSemaphore, 0, 0x7fffffff );
	devExt->WriteThreadStop = FALSE;
	devExt->WriteThread = NULL;

	/*
	   Initialize I/O Spin Locks
	*/
	KeInitializeSpinLock ( &devExt->InterruptLock );
	KeInitializeSpinLock ( &devExt->IntEventLock );
	KeInitializeSpinLock ( &devExt->ReadLock );
	KeInitializeSpinLock ( &devExt->ReadEventLock );

	/*
	   Initialize Events
	*/
	KeInitializeEvent ( &devExt->WriteIntEvent, SynchronizationEvent, FALSE );
	KeInitializeEvent ( &devExt->WriteStopEvent, SynchronizationEvent, FALSE );
	KeInitializeEvent ( &devExt->ReadStopEvent, SynchronizationEvent, FALSE );

	/*
	   set timeout to 10 seconds
	*/
	devExt->WriteTimeout = RtlConvertLongToLargeInteger(-1000000000);
	
	/*
	   set usb timeout to 5 seconds
	*/
	devExt->UsbTimeout = RtlConvertLongToLargeInteger(-500000000);

	/*
	   Initialize Fifo and Can state
	*/
	devExt->RXFifoSize = 0x400;
	devExt->RXFifo.msg = NULL;
	devExt->CanStarted = FALSE;

	/*
	   Mode
	*/
	/* WCHAR buffer[256];
	PHANDLE hKey;
	ntStatus 
		= IoOpenDeviceRegistryKey ( deviceObject, PLUGPLAY_REGKEY_DEVICE, READ_CONTROL, hKey );*/
	devExt->IoMode = EventMode;
	/*devExt->IoMode = PollMode; */
	devExt->IntSet = FALSE;
	devExt->PhysicalReconnectNeeded = FALSE;

	devExt->WaitForCompletionOnSuspend = FALSE;

	/*
		Create and Initialize Interrupt resources
		---------------------------------------------------------------------------------*/	
	devExt->IntPollPending = FALSE;
	devExt->ReadPollPending = FALSE;
	/*
	   allocate an irp for polling
	*/
	devExt->IntIrp
		= IoAllocateIrp ( devExt->NextDeviceObject->StackSize, FALSE );
	if ( !devExt->IntIrp )
	{
		KdPrint (( DRVNAME "FAILED to create IRP for polling the interrupt IN-Eendpoint\n" ));
		ntStatus = STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		/*
		allocate memory for the urb
		*/
		devExt->IntUrb
			= (PURB) ExAllocatePool ( NonPagedPool, sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER) );
		if ( !devExt->IntUrb )
		{
			KdPrint (( DRVNAME "FAILED to create URB for polling the interrupt IN-Endpoint\n" ));
			ntStatus = STATUS_INSUFFICIENT_RESOURCES;
		}
		else ntStatus = STATUS_SUCCESS;
	}
	if ( !NT_SUCCESS(ntStatus) )
	{
		KdPrint (( DRVNAME "FAILED to initialize Interrupt Resources; ntStatus(%X)\n", ntStatus ));

		Tc_DeleteInterruptResources ( devExt );
		IoDetachDevice ( devExt->NextDeviceObject );
		IoDeleteDevice ( deviceObject );
		return ntStatus;
	}

	/*
		Create and Initialize Bulk resources
		---------------------------------------------------------------------------------*/	
	devExt->ReadPending = FALSE;
	devExt->WritePending = FALSE;

	/*
	 *  allocate an irp for polling
	 */
	devExt->ReadPollIrp
		= IoAllocateIrp ( devExt->NextDeviceObject->StackSize, FALSE );
	if ( !devExt->ReadPollIrp )
	{
		KdPrint (( DRVNAME "FAILED to create IRP for polling the bulk IN-Endpoint\n" ));
		ntStatus = STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		/*
		*  allocate memory for the urb
		*/
		devExt->ReadPollUrb
			= (PURB) ExAllocatePool ( NonPagedPool, sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER) );
		if ( !devExt->ReadPollUrb )
		{
			KdPrint (( DRVNAME "FAILED to create URB for polling the bulk IN-Endpoint\n" ));
			ntStatus = STATUS_INSUFFICIENT_RESOURCES;
		}
		else ntStatus = STATUS_SUCCESS;
	}
	if ( !NT_SUCCESS(ntStatus) )
	{
		KdPrint (( DRVNAME "FAILED to initialize Bulk Resources; ntStatus(%X)\n", ntStatus ));

		Tc_DeleteInterruptResources ( devExt );
		Tc_DeleteBulkResources ( devExt );
		IoDetachDevice ( devExt->NextDeviceObject );
		IoDeleteDevice ( deviceObject );
		return ntStatus;
	}

	/*
		Initialize Firmware Resources
		---------------------------------------------------------------------------------*/
	devExt->FWBuffer = NULL;
	devExt->FWLength = 0;

	deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
	
	return ntStatus;
}


/******************************************************************************
	
	  Tc_StartDevice

 Description:
	Starts the device

  @devObject - the device Extension

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_StartDevice
(
 IN PDEVICE_OBJECT			  deviceObject
)
{
	/* Create an instance of URB */
	URB					urb;
	USHORT				vid;
	USHORT				pid;
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;

	/* create an instance of a configuration descriptor for the first step */
	USB_CONFIGURATION_DESCRIPTOR	ConfDesc4Size;
	PUSB_CONFIGURATION_DESCRIPTOR	ConfDesc;
	PUSBD_INTERFACE_LIST_ENTRY		interface_list;

	KdPrint (( DRVNAME "Tc_StartDevice\n" ));
	PAGED_CODE ();

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;
	
	/*---------------------------------------------------------------------------------------------
	*
	*   read DEVICE DECRIPTOR
	*
	*---------------------------------------------------------------------------------------------*/
	/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
	UsbBuildGetDescriptorRequest (
		&urb, 
		sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST), 
		USB_DEVICE_DESCRIPTOR_TYPE,
		0, 
		0, 
		&devExt->DescDevice, 
		NULL, 
		sizeof(devExt->DescDevice), 
		NULL );

	ntStatus 
		= Tc_CallToUSB ( deviceObject, &urb );

	if ( !NT_SUCCESS(ntStatus) )
	{
		KdPrint (( DRVNAME "FAILED to read device descriptor, ntStatus(%X)\n", ntStatus ));
		return ntStatus;
	}

	vid = devExt->DescDevice.idVendor;
	pid = devExt->DescDevice.idProduct;

	if ( vid == 0x0547 && pid == 0x2131 )
	{
		KdPrint (( DRVNAME "Start Firmware download\n" ));
		return Tc_LoadDevice ( deviceObject );		
	}
	else
	if ( vid == 0x1234 && pid == 0x5678 )
	{
		devExt->isFlashed = TRUE;
		KdPrint (( DRVNAME "USB-Tiny-CAN Product recognized\n" ));
	}
	else
	if ( vid == 0x1234 && pid == 0x5677 )
	{
		devExt->isFlashed = TRUE;
		KdPrint (( DRVNAME "USB-Tiny-CAN Prototype recognized\n" ));
	}
	else
	{
		KdPrint (( DRVNAME "USB-Tiny-CAN !!! NOT !!! recognized\n" ));
		return STATUS_UNSUCCESSFUL;
	}

	if ( devExt->DescDevice.bNumConfigurations != 1 || devExt->DescDevice.bcdUSB != 0x0110 )
	{
		KdPrint (( DRVNAME "FAILED invalid device descriptor. seems not to be the tiny-CAN device\n" ));
		return ntStatus;
	}

	/*
	KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
	KdPrint((DRVNAME "         Device Descriptor\n"));
	KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
	KdPrint((DRVNAME "bLength %.2x\n", devExt->DescDevice.bLength));
	KdPrint((DRVNAME "bDescriptorType %.2x\n", devExt->DescDevice.bDescriptorType));
	KdPrint((DRVNAME "bcdUSB %.4x\n", devExt->DescDevice.bcdUSB));
	KdPrint((DRVNAME "bDeviceClass %.2x\n", devExt->DescDevice.bDeviceClass));
	KdPrint((DRVNAME "bDeviceSubClass %.2x\n", devExt->DescDevice.bDeviceSubClass));
	KdPrint((DRVNAME "bDeviceProtocol %.2x\n", devExt->DescDevice.bDeviceProtocol));
	KdPrint((DRVNAME "bMaxPacketSize0 %.2x\n", devExt->DescDevice.bMaxPacketSize0));
	KdPrint((DRVNAME "bcdDevice %.4x\n", devExt->DescDevice.bcdDevice));
	KdPrint((DRVNAME "bNumConfigurations %.2x\n", devExt->DescDevice.bNumConfigurations));
	KdPrint((DRVNAME "iManufacturer %.4x\n", devExt->DescDevice.iManufacturer));
	KdPrint((DRVNAME "iProduct %.4x\n", devExt->DescDevice.iProduct));
	KdPrint((DRVNAME "iSerialNumber %.4x\n", devExt->DescDevice.iSerialNumber));
	KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
	*/
	

	/*--------------------------------------------------------------------------------------------------------------
	*
	*   read CONFIGURATION DECRIPTOR
	*
	*	To read the configuration descriptor requires two steps. The first one read in a configuration descriptor
	*	with a fixed size, from with we get the size of the complete configuration descriptor with all imbedded
	*	interface and endpoint descriptors
	*	The second step we read in with the same procedure supplied with memory for the complete descriptor
	*---------------------------------------------------------------------------------------------------------------*/

	/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
	UsbBuildGetDescriptorRequest (
		&urb, 
		sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST), 
		USB_CONFIGURATION_DESCRIPTOR_TYPE,
		0, 
		0, 
		&ConfDesc4Size, 
		NULL, 
		sizeof(ConfDesc4Size), 
		NULL );

	KdPrint (( DRVNAME "submit urb to get size of the complete Configuration Descriptor\n" ));
	
	ntStatus 
		= Tc_CallToUSB ( deviceObject, &urb );

	if ( !NT_SUCCESS(ntStatus) )
	{
		KdPrint (( DRVNAME "FAILED to read configuration descriptor 4s, ntStatus(%X)\n", ntStatus ));
		return ntStatus;
	}
	/* KdPrint (( DRVNAME "Length from 4s configuration descriptor %.4x\n",ConfDesc4Size.wTotalLength )); */


	/*--------------------------------------------------------------------------------------------------
	*
	*   read complete CONFIGURATION DECRIPTOR
	*
	*	allocate memory for the complete configuration, which we also store within the device extension
	*---------------------------------------------------------------------------------------------------*/
	ConfDesc 
		= (PUSB_CONFIGURATION_DESCRIPTOR) 
			ExAllocatePool ( NonPagedPool, ConfDesc4Size.wTotalLength );

	if ( !ConfDesc )
	{
		KdPrint (( DRVNAME "FAILED to allocate %X bytes for complete configuraton descriptor\n", ConfDesc4Size.wTotalLength ));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	/* first try for finaly free allocated configuration descriptor memory */
	__try
	{		
		PUSB_INTERFACE_DESCRIPTOR	InterfaceDesc;
		PURB						SelectUrb;
		UCHAR						Buffer[128];
		PUSBD_INTERFACE_INFORMATION InterfaceInfo;

		/* fill in the urb, now with memory for the complete conf. descriptor */
		/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
		UsbBuildGetDescriptorRequest (
			&urb, 
			sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST), 
			USB_CONFIGURATION_DESCRIPTOR_TYPE,
			0, 
			0, 
			ConfDesc, 
			NULL, 
			ConfDesc4Size.wTotalLength, 
			NULL );

		/* submit and verify */
		ntStatus 
			= Tc_CallToUSB ( deviceObject, &urb );

		if ( !NT_SUCCESS(ntStatus) )
		{
			KdPrint (( DRVNAME "FAILED to read complete configuration descriptor, ntStatus(%X)\n", ntStatus ));
			return ntStatus;
		}

		/*
		KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
		KdPrint((DRVNAME "         Configuration Descriptor\n"));
		KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
		KdPrint((DRVNAME "CD.bLength %.4x\n", ConfDesc->bLength));
		KdPrint((DRVNAME "CD.bNumInterfaces %.4x\n", ConfDesc->bNumInterfaces));
		KdPrint((DRVNAME "CD.bConfigurationValue %.4x\n", ConfDesc->bConfigurationValue));
		KdPrint((DRVNAME "CD.iConfiguration %.4x\n", ConfDesc->iConfiguration));
		KdPrint((DRVNAME "CD.bmAttributes %.4x\n", ConfDesc->bmAttributes));
		KdPrint((DRVNAME "CD.MaxPower %.4x\n", ConfDesc->MaxPower));
		KdPrint((DRVNAME "-----------------------------------------------------------------\n\n"));
		*/

		/*--------------------------------------------------------------------------------------------------
		*
		*   rextract INTERFACE DECRIPTOR
		*
		*	Locate the descriptor for the one and only interface we expect to find
		*	by means of bAlternateSetting you can switch beetween normal mode (0) and PeliCAN-Mode (1)
		*
		*	For easy search within the configuration descriptor,
		*	there are several routines supplied by the bus driver,
		*	which we use
		*---------------------------------------------------------------------------------------------------*/
		InterfaceDesc = 
			USBD_ParseConfigurationDescriptorEx (
				ConfDesc,	/* pointer to conf.desc. */
				ConfDesc,	/* start search at  */
				0x00,		/* bInterfaceNumber (-1 for any) */
				0x01,		/* bAlternateSetting: normal mode (0) and PeliCAN-Mode (1) */
				0xff,		/* bInterfaceClass - class to find (-1 for any) */
				0x00,		/* bInterfaceSubClass - subclass to find (-1 for any) */
				0xff		/* bInterfaceProtocoll - protocol to find (-1 for any) */
				);
		
		/* search succeesful ? */
		if ( !InterfaceDesc )
		{
			KdPrint (( DRVNAME "FAILED to find PeliCAN-Mode interface\n" ));
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}

		/*
		KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
		KdPrint((DRVNAME "         Interface Descriptor\n"));
		KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
		KdPrint((DRVNAME "ID.bLength %.4x\n",InterfaceDesc->bLength));
		KdPrint((DRVNAME "ID.bDescriptorType %.4x\n",InterfaceDesc->bDescriptorType));
		KdPrint((DRVNAME "ID.bInterfaceNumber %.4x\n",InterfaceDesc->bInterfaceNumber));
		KdPrint((DRVNAME "ID.bAlternateSetting %.4x\n",InterfaceDesc->bAlternateSetting));
		KdPrint((DRVNAME "ID.bNumEndpoints %.4x\n",InterfaceDesc->bNumEndpoints));
		KdPrint((DRVNAME "ID.bInterfaceClass %.4x\n",InterfaceDesc->bInterfaceClass));
		KdPrint((DRVNAME "ID.bInterfaceSubClass %.4x\n",InterfaceDesc->bInterfaceSubClass));
		KdPrint((DRVNAME "ID.bInterfaceProtocol %.4x\n",InterfaceDesc->bInterfaceProtocol));
		KdPrint((DRVNAME "ID.iInterface %.4x\n",InterfaceDesc->iInterface));
		KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
		*/

		/*--------------------------------------------------------------------------------------------------
		*
		*   select INTERFACE
		*
		*	In order to get the pipes, we're interested in, we must create a urb to select the PeliCAN-Interface (1)
		*	This urb is created by USBD_CreateConfigurationRequestEx, to which we give
		*	an array of USBD_INTERFACE_LIST_ENTRY entrys
		*
		*	USBD_INTERFACE_LIST_ENTRY has two members.
		*	the first is a pointer to the interface descriptor
		*	the second one will be filled by the bus driver with some information from which we get the
		*	for us interesting pipes
		*	The second one is a structure with an arraytype member Pipes of USBD_PIPE_INFORMATION
		*---------------------------------------------------------------------------------------------------*/
		interface_list = 
			ExAllocatePool ( NonPagedPool, sizeof(USBD_INTERFACE_LIST_ENTRY) * 2 );

		if ( !interface_list )
		{
			KdPrint (( DRVNAME "FAILED to find allocate memory for the interface list\n" ));
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}
		interface_list[0].InterfaceDescriptor = InterfaceDesc;
		interface_list[0].Interface = NULL;
		interface_list[1].InterfaceDescriptor = NULL;
		interface_list[1].Interface = NULL;

		SelectUrb 
			= USBD_CreateConfigurationRequestEx ( ConfDesc, interface_list );

		if ( !SelectUrb )
		{
			KdPrint (( DRVNAME "FAILED to create configuration request\n" ));
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		
		__try
		{
			/* Verify that the interface describes exactly the endpoints we expect */
			if ( InterfaceDesc->bNumEndpoints != 0x03 )
			{
				KdPrint (( DRVNAME "FAILED: %d endpoints; seems not to be the tiny-CAN device\n", InterfaceDesc->bNumEndpoints ));
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}

			InterfaceInfo = interface_list[0].Interface;

			/* Verify that the interface describes exactly 3 Pipes we expect */
			if ( InterfaceInfo->NumberOfPipes != 0x03 )
			{
				KdPrint (( DRVNAME "FAILED: %d pipes; seems not to be the tiny-CAN device\n", InterfaceDesc->bNumEndpoints ));
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}

			/* not needed, but who knows ... */
			ASSERT ( InterfaceInfo->NumberOfPipes == InterfaceDesc->bNumEndpoints );

			/*--------------------------------------------------------------------------------------------------
			*
			*    to influence the host controller drivers scheduling algorithm, uncomment the following lines
			*
			*	for (int i=0; i<InterfaceInfo->NumberOfPipes; ++i)
			*	InterfaceInfo->Pipes[i].MaximumTransferSize = USBD_DEFAULT_MAXIMUM_TRANSFER_SIZE;
			*---------------------------------------------------------------------------------------------------*/

			/*--------------------------------------------------------------------------------------------------
			*
			*    Submit the set-configuration request
			*
			*---------------------------------------------------------------------------------------------------*/
			KdPrint (( DRVNAME "Submit URB for selecting the configuration\n" ));
			
			ntStatus 
				= Tc_CallToUSB ( deviceObject, SelectUrb );

			if ( !NT_SUCCESS(ntStatus) )
			{
				KdPrint (( DRVNAME "FAILED to select configuration, ntStatus(%X)\n", ntStatus ));
				return ntStatus;
			}

			/*
			for ( ULONG i=0; i<InterfaceInfo->NumberOfPipes; i++ )
			{
				KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
				KdPrint((DRVNAME "         Pipe %i Information (Endpoints)\n", i));
				KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
				KdPrint((DRVNAME "EP%i.PipeType  %.4x\n", i, InterfaceInfo->Pipes[i].PipeType));
				KdPrint((DRVNAME "EP%i.MaximumPacketSize  %.4x\n", i, InterfaceInfo->Pipes[i].MaximumPacketSize));
				KdPrint((DRVNAME "EP%i.EndpointAddress  %.4x\n", i, InterfaceInfo->Pipes[i].EndpointAddress));
				KdPrint((DRVNAME "EP%i.MaximumTransferSize  %.4x\n", i, InterfaceInfo->Pipes[i].MaximumTransferSize));
				KdPrint((DRVNAME "EP%i.Interval  %.4x\n", i, InterfaceInfo->Pipes[i].Interval));
				KdPrint((DRVNAME "-----------------------------------------------------------------\n"));
			}
			*/

			/*--------------------------------------------------------------------------------------------------
			*
			*    check the endpoints to ensure the device we manage is a tiny-CAN type
			*
			*---------------------------------------------------------------------------------------------------*/
			/* 
				If0, Alt1, EndPoint0
				0x02, // bmAttributes (bulk endpoint)
				0x02, // bEndpointAddress (OUT2)
				0x40, 0x00, // wMaxPacketSize
				0x00, // bInterval (0ms)			
			*/
			if ( InterfaceInfo->Pipes[0].PipeType != 0x02 ||
				InterfaceInfo->Pipes[0].EndpointAddress != 0x02 ||
				InterfaceInfo->Pipes[0].MaximumPacketSize != 0x0040 ||
				InterfaceInfo->Pipes[0].Interval != 0x00 )
			{
				KdPrint (( DRVNAME "FAILED: pipe0 (Bulk OUT) invalid; seems not to be the tiny-CAN device\n" ));
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}
			/* 
				If0, Alt1, EndPoint1
				0x02, // bmAttributes (bulk endpoint)
				0x82, // bEndpointAddress (IN2)
				0x40, 0x00, // wMaxPacketSize
				0x00, // bInterval (0ms)			
			*/
			if ( InterfaceInfo->Pipes[1].PipeType != 0x02 ||
				InterfaceInfo->Pipes[1].EndpointAddress != 0x82 ||
				InterfaceInfo->Pipes[1].MaximumPacketSize != 0x0040 ||
				InterfaceInfo->Pipes[1].Interval != 0x00 )
			{
				KdPrint(( DRVNAME "FAILED: pipe1 (Bulk IN) invalid; seems not to be the tiny-CAN device\n" ));
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}
			/* 
				If0, Alt1, EndPoint3
				0x03, // bmAttributes (interrupt endpoint)
				0x84, // bEndpointAddress (IN4)
				0x10, 0x00, // wMaxPacketSize
				0x01, // bInterval (1ms)			
			*/
			if ( InterfaceInfo->Pipes[2].PipeType != 0x03 ||
				InterfaceInfo->Pipes[2].EndpointAddress != 0x84 ||
				InterfaceInfo->Pipes[2].MaximumPacketSize != 0x0010 ||
				InterfaceInfo->Pipes[2].Interval != 0x01 )
			{
				KdPrint (( DRVNAME "FAILED: pipe2 (Interrupt IN) invalid; seems not to be the tiny-CAN device\n" ));
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}

			/*--------------------------------------------------------------------------------------------------
			*
			*    save the configuration and pipe handles
			*
			*---------------------------------------------------------------------------------------------------*/
			devExt->ConfPipe = SelectUrb->UrbSelectConfiguration.ConfigurationHandle;
			devExt->BulkOUTPipe = InterfaceInfo->Pipes[0].PipeHandle;
			devExt->BulkINPipe = InterfaceInfo->Pipes[1].PipeHandle;
			devExt->InterruptINPipe = InterfaceInfo->Pipes[2].PipeHandle;
			
			/*--------------------------------------------------------------------------------------------------
			*
			*    save the configuration descriptor
			*
			*---------------------------------------------------------------------------------------------------*/
			devExt->ConfDesc = ConfDesc;
			ConfDesc = NULL;

			devExt->IoQueueState = InActive;

			/*--------------------------- MAKRO BUILD OF URB ---------------------------*/
			UsbBuildGetDescriptorRequest (
				&urb,
				sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
				USB_STRING_DESCRIPTOR_TYPE,
				0,
				0,
				Buffer,
				NULL,
				sizeof(Buffer),
				NULL );

			ntStatus
				= Tc_CallToUSB ( deviceObject, &urb );

			if ( !NT_SUCCESS(ntStatus) )
				devExt->DefaultLanguage = 0x0409;
			else
				devExt->DefaultLanguage = *(PSHORT)( Buffer + 2 );
		}
		__finally
		{
			ExFreePool ( SelectUrb );
		}

	}
	__finally
	{
		if ( ConfDesc )
			ExFreePool ( ConfDesc );
		if ( interface_list )
			ExFreePool ( interface_list );
	}

	return STATUS_SUCCESS;
}


/******************************************************************************
 
   Tc_StopDevice

 Description:
	

  @deviceObject - the device Extension
  @DeConfigureDevice

 no return value

******************************************************************************/
VOID Tc_StopDevice
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN BOOLEAN			DeConfigureDevice
)
{
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;

	KdPrint (( DRVNAME "Tc_StopDevice() %i\n", KeGetCurrentIrql() ));
	PAGED_CODE ();

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;
	
	/* !!! NEW !!! must be called at PASSIVE_LEVEL
	   available for w2k3
	   KeFlushQueuedDPCs();
	 */
	
	/*
	   notify a connected application the remove event
	*/
	if ( devExt->IntEvent )
	{
		KLOCK_QUEUE_HANDLE hLock;
		KeAcquireInStackQueuedSpinLock ( &devExt->IntEventLock, &hLock );
		/*----------------------------------------------------------------*/							
			RtlZeroMemory ( &devExt->IntEventBuffer, sizeof(UCI_INT_DATA) );
			devExt->IntEventBuffer.st_code = 0xc0;
			devExt->IntSet = TRUE;
			if ( devExt->IoMode )
			{
				if (devExt->IntEvent)
					KeSetEvent(devExt->IntEvent, EVENT_INCREMENT, FALSE);
			}
			else
			{
				/* Start a dpc for processing an interrupt Irp
				   check IRQ-Level
				   --> we can also do this without a dpc
			     */
				KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) FALSE );
			}
		/*----------------------------------------------------------------*/
		KeReleaseInStackQueuedSpinLock ( &hLock );
	}

	if ( DeConfigureDevice )
	{
		URB urb;
		UsbBuildSelectConfigurationRequest (
			&urb, 
			sizeof(struct _URB_SELECT_CONFIGURATION), 
			NULL );

		ntStatus 
			= Tc_CallToUSB ( deviceObject, &urb );
	}

	if ( devExt->ConfDesc )
		ExFreePool ( devExt->ConfDesc );

	devExt->ConfDesc = NULL;
	devExt->BulkINPipe = NULL;
	devExt->BulkOUTPipe = NULL;
	devExt->ConfPipe = NULL;
}


/******************************************************************************
 
   Tc_DriverUnload

 Description:
	nothing to do, but the debug message is needed to verify test scenarios

  @driverObject - the driver object

 not return value

******************************************************************************/
VOID Tc_DriverUnload
(
 IN PDRIVER_OBJECT driverObject
)
{
	KdPrint (( DRVNAME "Tc_DriverUnload\n" ));
	PAGED_CODE ();
	return;
}


/******************************************************************************
 
   Tc_DispatchInternalControl

 Description:
	

  @deviceObject - the device object
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_DispatchInternalControl
(
 IN PDEVICE_OBJECT  deviceObject, 
 IN PIRP			Irp
)
{
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;

	KdPrint (( DRVNAME "Tc_DispatchInternalControl\n" ));
	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	ntStatus 
		= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );

	if ( !NT_SUCCESS(ntStatus) )
		Tc_FinishIrp2 ( Irp, ntStatus );
	else
	{
		IoSkipCurrentIrpStackLocation ( Irp );

		ntStatus 
			= IoCallDriver ( devExt->NextDeviceObject, Irp );

		IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
	}
	return ntStatus;
}


/******************************************************************************
	
	  Tc_RemoveDevice

 Description:
	Detach the DeviceObject and Delete it.

  @deviceObject - the device object

 not returns value

******************************************************************************/
VOID Tc_RemoveDevice
(
 IN PDEVICE_OBJECT deviceObject
)
{
	PDEVICE_EXTENSION devExt;

	KdPrint (( DRVNAME "Tc_RemoveDevice\n" ));
	PAGED_CODE();
	
	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;	
	
	if ( devExt->ReadEvent )
	{
		ObDereferenceObject ( devExt->ReadEvent );
		devExt->ReadEvent = NULL;
	}

	if ( devExt->IntEvent )
	{
		ObDereferenceObject ( devExt->IntEvent );
		devExt->IntEvent = NULL;
	}

	if ( devExt->IntIrp )
		Tc_DeleteInterruptResources ( devExt );

	if ( devExt->ReadPollIrp )
		Tc_DeleteBulkResources ( devExt );

	if ( devExt->FWBuffer != NULL )
		ExFreePool ( devExt->FWBuffer );

	if ( devExt->NextDeviceObject )
		IoDetachDevice ( devExt->NextDeviceObject );
	
	devExt->Pdo = NULL;
	IoDeleteDevice ( devExt->DeviceObject );

}

/******************************************************************************
 
   Tc_DispatchCreate

 Description: 
	This routine is called each time a CreateHandle opens a handle.
	To provide a controlling mechanism we increment the opened handles.
	The tiny CAN Device and it's interrupt and read use is exclusive and
	not shared. The Write operation is able to share.
	Hence we allow only one handle to be opened.

 Issue:
	We must provide some mechanism for handles which were not closed, so they
	block other applications to use the can device.

  @deviceObject - the device object
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_DispatchCreate
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
)
{
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	ioStack ;

	KdPrint (( DRVNAME "Tc_DispatchCreate\n" ));
	PAGED_CODE ();
	
	ioStack 
		= IoGetCurrentIrpStackLocation ( Irp );
	
	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;
	
	ntStatus = STATUS_UNSUCCESSFUL;

	/*
	   maybe verify the pnp state
	   it's omitted because of maintenance reasons
	   in a release version this should be build in
	*/
	if ( devExt->openHandles < 1000 && devExt->isFlashed )
	{
		ntStatus = IoAcquireRemoveLock ( &devExt->RemoveLock, ioStack->FileObject );

		if ( NT_SUCCESS(ntStatus) )
			InterlockedIncrement ( &devExt->openHandles );
	}

	return Tc_FinishIrp2 ( Irp, ntStatus );
}


/******************************************************************************
 
   Tc_DispatchClose

 Description:
	

  @deviceObject - the device object
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_DispatchClose
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
)
{
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	ioStack;

	PAGED_CODE();
	KdPrint (( DRVNAME "Tc_DispatchClose\n" ));

	ioStack 
		= IoGetCurrentIrpStackLocation ( Irp );

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	InterlockedDecrement ( &devExt->openHandles );
		
	IoReleaseRemoveLock ( &devExt->RemoveLock, ioStack->FileObject );

	if ( devExt->openHandles == 0 && devExt->CanStarted )
		Tc_StopCan ( deviceObject, !( devExt->DevState == SurpriseRemoved || devExt->DevState == Removed || devExt->PhysicalReconnectNeeded ) );

	return Tc_FinishIrp1 ( Irp );
}


/******************************************************************************
 
   Tc_DispatchCleanup

 Description:
	Clean up all user context resources if allocated and cancel all pending
	irps in the queue belonging to the user handle
	This routine is called before IRP_MJ_CLOSE is send.

 Note:
	this is a device for exclusive use, that is, it is safe to cancel all
	outstanding irps and dereference all event objects
	For further implementations, if shared use is intended, which only the
	WriteFile operation does make sense, this has to be changed !!!
	ReadFile and Interrupt handling for shared use implicates much code
	to handle multiple events and lists to manage them.

  @deviceObject - the device object
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_DispatchCleanup
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
)
{
	PIRP				pirp;
	QUEUECONTEXT		PeekContext;
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	ioStack;

	KdPrint (( DRVNAME "Tc_DispatchCleanup\n" ));
	PAGED_CODE ();

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	ioStack 
		= IoGetCurrentIrpStackLocation ( Irp );

		/*	could cause page faults !!! */

	if ( devExt->ReadEvent && devExt->ReadFileObject == ioStack->FileObject )
	{
		ObDereferenceObject ( devExt->ReadEvent );
		devExt->ReadEvent = NULL;
		devExt->ReadFileObject = NULL;
	}

	if ( devExt->IntEvent && devExt->IntFileObject == ioStack->FileObject )
	{
		ObDereferenceObject ( devExt->IntEvent );
		devExt->IntEvent = NULL;
		devExt->IntFileObject = NULL;
	}

	PeekContext.Context = &devExt->NoMatter;
	PeekContext.FileObject = ioStack->FileObject;

	pirp = NULL;
	while ( pirp )
	{
		pirp = IoCsqRemoveNextIrp( &devExt->IoQueue, &PeekContext );
		Tc_FinishIrp2 ( pirp, STATUS_CANCELLED );
	}

	return Tc_FinishIrp1 ( Irp );
}