/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Declarations for USB-Tiny-CAN driver
 *		used c++ comments for vs.net integration
 *
 * TODO:
 *
 * History of changes:
 *	 2005/03/15 - 0.1 - first version
 *
 * Thesis: 
 *		Treiberentwicklung fr MS Windows Systeme am
 *		Beispiel des Windows Driver Models fr die Tiny CAN USB-Schnittstelle.
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef TCAN_USB_H
#define TCAN_USB_H

#define XPERF

#ifdef XPERF
	#include <ntddk.h>
#else
	#include <wdm.h>
	#include <csq.h>
#endif

#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "tcan_usb_global.h"

#define DRVNAME					"[tcan_usb]> "
#define DRVERSION				0x00020104

/*
 * CAN and only driver related stuff
*/
#define CANMSG_BUF_EMPTY		0x00	/* fifo is empty */
#define CANMSG_BUF_PEND			0x01	/* message(s) pending */
#define CANMSG_BUF_FULL			0x02	/* fifo is full */
#define CANMSG_BUF_OVR			0x04	/* fifo overrun */

#define	UCI_ST_CAN_RXFIFO_OVR	0x01
#define	UCI_ST_CAN_ERROR_INT	0x02
#define	UCI_ST_FW_RXFIFO_OVR	0x04
#define	UCI_ST_TX_MSG			0x08
#define	UCI_ST_TX_MSG_CNT		0xf0
#define UCI_FAIL_BEHAVE_COUNT	16

/* some IDs of USB-CAN-Interface Firmware (UCI Firmware)
 * Vendor ID here is temporary,
 * you have to get VENDOR_ID from USB-Implementers-Forum */
#define UCI_VENDOR_ID			0x1234	/* ID of USBCAN device Vendor*/
#define UCI_PROTOTYPE_ID		0x5677	/* ID of USBCAN-Prototype*/
#define UCI_PRODUCT_ID			0x5678	/* Product ID of USB-Tiny-CAN*/
#define UCI_DEV_RELEASE			0x0001	/* Release #*/

/* Our own vendor requests (bRequest) */
#define UCI_CAN_CHIP_ACCESS		0xB0
#define UCI_CAN_SEND_MSG		0xB1  /*not implemented yet, do we need it?*/
#define UCI_CAN_RECEIVE_MSG		0xB2  /*not implemented yet*/

/* parameters (wIndexL) for vendor request UCI_CAN_CHIP_ACCESS */
#define UCI_READ_CAN_REGISTER	0x01
#define UCI_WRITE_CAN_REGISTER	0x02
#define UCI_GET_CAN_REGISTERS	0x03
#define UCI_SET_BUS_TIMINGS		0x04
#define UCI_SET_ACCEPT_CODE		0x05
#define UCI_SET_ACCEPT_MASK		0x06
#define UCI_CAN_RESET			0x07
#define UCI_CAN_START			0x08
#define UCI_CAN_STOP			0x09

/* Vendor specific requests of avrpipe firmware */
#define	RQ_AVR_DATA_AVAILABLE	0xB0
#define	RQ_GET_AVR_STATUS		0xB1
#define	RQ_SEND_TO_AVR			0xB2
#define	RQ_READ_FROM_AVR		0xB3
#define	RQ_READ_AVR_SRAM		0xB4

/* AN2131SC parameter and vendor reguests */
#define EZUSB_RAM_SIZE			0x2000
#define FIRMWARE_LOAD			0xA0
#define VENDOR_RQ_OUT			0x40
#define VENDOR_RQ_IN			0xC0

typedef 
struct _uci_canmsg 
{
	UCHAR		flags; /* (std/ext,rtr,dlc) as in Frame Info Reg */
	/* ID as it appears in chip registers */
	UCHAR		id1;
	UCHAR		id2;
	UCHAR		id3;
	UCHAR		id4;
    UCHAR		msgdata[CAN_MSG_LENGTH];
} 
UCI_CANMSG, *PUCI_CANMSG;

#define BULK_MSG_COUNT		4

typedef
struct _bulk_buffer
{
	UCHAR		mcount;
	UCI_CANMSG	message1;
	UCI_CANMSG	message2;
	UCI_CANMSG	message3;
	UCI_CANMSG	message4;
	UCHAR		reserved[11];
} 
BULK_BUFFER, *PBULK_BUFFER;

typedef
struct canmsg_fifo 
{
	volatile ULONG	head;
	volatile ULONG	tail;
	volatile UCHAR	status;
	PCANMSG			msg;
}
BULK_FIFO, *PBULK_FIFO;

typedef
enum canstate 
{
    NeverStarted,
    Stopped,
    Working,
    PendingStop,
    PendingRemove,
    SurpriseRemoved,
    Removed
}
CANSTATE;

typedef
enum queuestate 
{
    Active,
    Stalled,
    InActive
}
QUEUESTATE;

typedef
struct _QueueContext
{
	PIO_CSQ_IRP_CONTEXT	Context;
	PFILE_OBJECT		FileObject;
}
QUEUECONTEXT, *PQUEUECONTEXT;

typedef
struct _STOP_WORKITEM
{
	PDEVICE_OBJECT		deviceObject;
	PIO_WORKITEM		wi;
} STOP_WORKITEM, *PSTOP_WORKITEM;

/*----------------------------------------------------------------
  Device extension structure
  ----------------------------------------------------------------*/
typedef 
struct tagDEVICE_EXTENSION 
{
	PDEVICE_OBJECT			DeviceObject;
	PDEVICE_OBJECT			Pdo;	
	PDEVICE_OBJECT			NextDeviceObject;

	IO_REMOVE_LOCK			RemoveLock;
	DEVICE_CAPABILITIES		CanCapabilities;
	DEVICE_POWER_STATE		DevPower;
	DEVICE_POWER_STATE		DevPowerNext;

	ULONG					FWLength;
	PCHAR					FWBuffer;

	// state of can device, TRUE if started, FALSE if stopped
	BOOLEAN					CanStarted;		
	KSPIN_LOCK				CanStartedLock;
	CIOMODE					IoMode;
	// TRUE = Firmware flashed, otherwise it's not the correct device
	BOOLEAN					isFlashed;
	BOOLEAN					PhysicalReconnectNeeded;

	// Plug and Play aware state
	CANSTATE				DevState;
	// previous pnp state
	CANSTATE				DevStatePrev;
	// spin lock for pnp state
	KSPIN_LOCK				DevStateLock;

	// DPC object for PollMode Read/Interrupt requests
	KDPC					IoDpc;

	// Implementing Cancel Safe Queues, for win9x one must build against csq.h, and deliver csq.dll
	IO_CSQ					IoQueue;
	// State of the Queue if stalled
	QUEUESTATE				IoQueueState;
	// spin lock for queue state
	KSPIN_LOCK				IoQueueStateLock;
	// event for events to stop
	IO_CSQ_IRP_CONTEXT		Read;
	IO_CSQ_IRP_CONTEXT		Write;
	IO_CSQ_IRP_CONTEXT		Interrupt;
	IO_CSQ_IRP_CONTEXT		NoMatter;

	// List to queue the pending Irps
	LIST_ENTRY				IoQueueHead;
	// SpinLock for access to the list
	KSPIN_LOCK				IoQueueLock;
	// to count the open handles
	LONG					openHandles;			
	
	USBD_CONFIGURATION_HANDLE		ConfPipe;
	USB_DEVICE_DESCRIPTOR			DescDevice;
	PUSB_CONFIGURATION_DESCRIPTOR	ConfDesc;
	SHORT							DefaultLanguage;

	LARGE_INTEGER	UsbTimeout;

	/*----------------------------------------------------------------
	   Bulk Pipes
	  ----------------------------------------------------------------*/
	USBD_PIPE_HANDLE BulkINPipe;
	USBD_PIPE_HANDLE BulkOUTPipe;

	BOOLEAN			ReadPollPending;
	// this spin lock locks the read pending flag
	KSPIN_LOCK		ReadPollLock;
	BULK_BUFFER		ReadPollBuffer;
	PIRP			ReadPollIrp;
	PURB			ReadPollUrb;

	BOOLEAN			ReadPending;
	KSPIN_LOCK		ReadLock;
	PKEVENT			ReadEvent;
	// we need this spin lock, to prevent something from dereference the event, while we are signaling it
	KSPIN_LOCK		ReadEventLock;
	PFILE_OBJECT	ReadFileObject;
	LONG			ReadCount;
	KEVENT			ReadStopEvent;

	// Receive
	BULK_FIFO		RXFifo;
	// this spin lock locks the receive fifo
	KSPIN_LOCK		RXFifoLock;
	LONG			RXFifoOverrun;
	ULONG			RXFifoSize;

	// Transmit
	LARGE_INTEGER	WriteTimeout;
	KEVENT			WriteIntEvent;
	UCI_INT_DATA	WriteIntBuffer;
	// this spin lock locks the write pending flag
	KSPIN_LOCK		WriteLock;
	BOOLEAN			WritePending;
	KEVENT			WriteStopEvent;

    PETHREAD		WriteThread;
	BOOLEAN			WriteThreadStop;
    KSEMAPHORE		WriteThreadSemaphore;


	
	/*----------------------------------------------------------------
	   Interrupt Pipe
	  ----------------------------------------------------------------*/
	// handle to the interrupt IN-Endpoint
	USBD_PIPE_HANDLE	InterruptINPipe;	

	// IRP used to poll for interrupts
	PIRP				IntIrp;
	// URB used to poll for interrupts
	PURB				IntUrb;	
	// buffer for the Urb to store the data
	UCI_INT_DATA		IntBuffer;
	// polling irp is pending
	BOOLEAN				IntPollPending;
	// lock for managing polling IRP
	KSPIN_LOCK			InterruptLock;		

	PKEVENT				IntEvent;
	// we need this spin lock, to prevent something from dereference the event, while we are signaling it
	KSPIN_LOCK			IntEventLock;
	PFILE_OBJECT		IntFileObject;
	// if we have an interrupt event for the user, this is the Buffer to store it
	UCI_INT_DATA		IntEventBuffer;
	// Interrupt signal to reset
	BOOLEAN				IntSet;
	
	LONG numFWOvr;
	LONG cacheFWOvr;

	// Values should moved to the Registry
	BOOLEAN				WaitForCompletionOnSuspend;

}
DEVICE_EXTENSION, *PDEVICE_EXTENSION;



/*----------------------------------------------------------------
   Driver entry and management prototypes
  ----------------------------------------------------------------*/
NTSTATUS Tc_AddDevice
(
 IN PDRIVER_OBJECT driverObject, 
 IN PDEVICE_OBJECT devObject
);

VOID Tc_DriverUnload
(
 IN PDRIVER_OBJECT driverObject
);


/*----------------------------------------------------------------
   Common prototypes
  ----------------------------------------------------------------*/
NTSTATUS Tc_CallToUSBExt
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PURB			urb,
 IN BOOLEAN			wait
);

NTSTATUS Tc_CallToUSB
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PURB			urb
);

NTSTATUS Tc_ConjureDescriptorStringToBuffer
(
 PDEVICE_OBJECT	deviceObject,
 UCHAR			index,
 PIRP			Irp,
 LONG			outLength
);

NTSTATUS Tc_FinishIrp
(
 IN PIRP		Irp, 
 IN NTSTATUS	ntStatus, 
 IN ULONG_PTR	Info
);

NTSTATUS Tc_FinishIrp2
(
 IN PIRP		Irp, 
 IN NTSTATUS	ntStatus
);

NTSTATUS Tc_FinishIrp1
(
 IN PIRP		Irp
);

NTSTATUS Tc_FinishIrp0
(
 IN PIRP		Irp, 
 IN NTSTATUS	ntStatus
);

/*----------------------------------------------------------------
   Global functions
  ----------------------------------------------------------------*/
NTSTATUS Tc_LoadDevice
(
 IN PDEVICE_OBJECT			  deviceObject
);

CHAR GetHex
( 
 IN CHAR c 
);

NTSTATUS Tc_StartPolling
(
 IN PDEVICE_EXTENSION	devExt,
 IN BOOLEAN				isInterrupt
);

VOID Tc_StopPolling
(
 IN PDEVICE_EXTENSION	devExt,
 IN BOOLEAN				isInterrupt
);

VOID Tc_RemoveDevice
(
 IN PDEVICE_OBJECT deviceObject
);

NTSTATUS Tc_StartDevice
(
 IN PDEVICE_OBJECT deviceObject
);

VOID Tc_StopDevice
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN BOOLEAN			DeConfigureDevice
);

VOID Tc_AcquireLock
(
 IN  PIO_CSQ	pQueue,
 OUT PKIRQL		Irql
);

VOID Tc_ReleaseLock
(
 IN  PIO_CSQ	pQueue,
 OUT KIRQL		Irql
);

VOID Tc_InsertIrp
(
 IN PIO_CSQ pQueue,
 IN PIRP	Irp
);

PIRP Tc_PeekNextIrp
(
 IN PIO_CSQ pQueue,
 IN PIRP	Irp,
 IN PVOID	PeekContext
);

VOID Tc_RemoveIrp
(
 IN PIO_CSQ	pQueue,
 IN PIRP	Irp
);

VOID Tc_CompleteCanceledIrp
(
 IN PIO_CSQ	pQueue,
 IN PIRP	Irp
);

VOID Tc_SetPnpState
(
 IN PDEVICE_EXTENSION	devExt,
 IN CANSTATE			pnpState
);

/*----------------------------------------------------------------
   Bulk prototypes
  ----------------------------------------------------------------*/
VOID Tc_DeleteBulkResources
(
 IN PDEVICE_EXTENSION devExt
);

NTSTATUS Tc_OnBulkRead
(
 IN PDEVICE_OBJECT		deviceObject, 
 IN PIRP				Irp, 
 IN PDEVICE_EXTENSION	devExt
);

VOID Tc_StartIoDpc
(
 IN PKDPC				Dpc,
 IN	PDEVICE_OBJECT		deviceObject,
 IN	PIRP				Irp,
 IN BOOLEAN				isRead
);

VOID Tc_StartReadIo
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);

VOID Tc_StartWriteIoThread
(
 IN PVOID Context
);


/*----------------------------------------------------------------
   Interrpt prototypes
  ----------------------------------------------------------------*/
VOID Tc_DeleteInterruptResources
(
 IN PDEVICE_EXTENSION	devExt
);

NTSTATUS Tc_OnInterrupt
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp, 
 PDEVICE_EXTENSION	devExt
);

VOID WAStop_Device
(
 IN PSTOP_WORKITEM wi
);

VOID Tc_StartInterruptIo
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);

/*----------------------------------------------------------------
   I/O request prototypes
  ----------------------------------------------------------------*/
NTSTATUS Tc_DispatchCreate
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
);

NTSTATUS Tc_DispatchClose
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
);

NTSTATUS Tc_DispatchCleanup
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
);

NTSTATUS Tc_DispatchReadWrite
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
);

NTSTATUS Tc_DispatchControl
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
);

NTSTATUS Tc_DispatchInternalControl
(
 IN PDEVICE_OBJECT  deviceObject, 
 IN PIRP			Irp
);

NTSTATUS Tc_DispatchWmi
(
 IN PDEVICE_OBJECT deviceObject,
 IN PIRP Irp
);

NTSTATUS Tc_InitAndStartCan
(
 IN PDEVICE_EXTENSION	devExt,
 IN PUCHAR				ubuffer
);

NTSTATUS Tc_StopCan
(
 IN PDEVICE_OBJECT	deviceObject,
 IN BOOLEAN			CanControl
);

NTSTATUS Tc_StopCanDL
(
 IN PDEVICE_EXTENSION	devExt
);

/*----------------------------------------------------------------
   Plug and Play prototypes
  ----------------------------------------------------------------*/
NTSTATUS Tc_DispatchPnp
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
);
NTSTATUS Tc_DispatchPower
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);

NTSTATUS Tc_IrpSetEventCR
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp,
 IN PVOID			Context
);

VOID Tc_StallQueueAwait
(
 IN PDEVICE_EXTENSION	devExt
);

VOID Tc_StopQueueAwait
(
 IN PDEVICE_EXTENSION	devExt
);

VOID Tc_RestartQueue
(
 IN PDEVICE_OBJECT	deviceObject
);

VOID Tc_ClearIoQueue
(
 IN PDEVICE_EXTENSION	devExt
);

#endif
